from nawah import __version__

from typing import Literal

import argparse, os, sys, logging, asyncio, datetime, traceback

from ._logging import logger, handler, formatter


def launch(
	args: argparse.Namespace,
	custom_launch: Literal['test', 'generate_ref', 'generate_models'] = None,
):

	# [DOC] Update Config with Nawah CLI args
	from nawah.config import Config
	from nawah.utils import _process_config

	Config._nawah_version = __version__
	if custom_launch not in ['generate_ref', 'generate_models']:
		Config.env = args.env
	if not custom_launch:
		Config.force_admin_check = args.force_admin_check

	# [DOC] Check for debug CLI Arg
	if args.debug:
		Config.debug = True
		logger.setLevel(logging.DEBUG)
	# [DOC] Check for log CLI Arg
	if not custom_launch and args.log:
		logger.removeHandler(handler)
		if not os.path.exists(os.path.join('.', 'logs')):
			os.makedirs(os.path.join('.', 'logs'))
		file_handler = logging.FileHandler(
			filename=os.path.join(
				'.',
				'logs',
				f'{datetime.datetime.utcnow().strftime("%d-%b-%Y")}.log',
			)
		)
		file_handler.setFormatter(formatter)
		logger.addHandler(file_handler)
		logger.setLevel(logging.DEBUG)

	from nawah.app import run_app
	import json

	try:
		try:
			sys.path.append('.')
			nawah_app = __import__('nawah_app')
			app_config = nawah_app.config

			with open('packages.json') as f:
				packages = json.loads(f.read())

			Config._app_packages = {k: v['version'] for k, v in packages.items()}
		except ModuleNotFoundError:
			logger.error(f'No \'nawah_app.py\' file found in CWD. Exiting.')
			exit(1)
		except AttributeError:
			logger.error(
				f'File \'nawah_app.py\' was found but it doesn\'t have \'config\' method. Exiting.'
			)
			exit(1)
		logger.info(
			f'Found app \'{app_config.name} (v{app_config.version})\'. Attempting to load App Config.'
		)
		Config._app_name = app_config.name
		Config._app_version = app_config.version
		Config._app_default_package = app_config.default_package

		# [DOC] Check all required values of App Config are present
		if not Config._app_name or not Config._app_version or not Config._app_default_package:
			logger.error(
				f'App Config are missing at least one of the following \'name, version, default_package\'. Exiting.'
			)
			exit(1)

		Config._app_path = os.path.realpath('.')
		# [DOC] Read app_config and update Config accordingly
		# [DOC] Check envs, env
		if custom_launch not in ['generate_ref', 'generate_models'] and app_config.envs:
			if not args.env and not app_config.env:
				logger.error(
					'App Config Attr \'envs\' found, but no \'env\' App Config Attr, or CLI Attr were defined.'
				)
				exit(1)
			if args.env:
				if args.env in app_config.envs.keys():
					logger.info(f'Setting \'env\' Config Attr to \'env\' CLI Arg value \'{args.env}\'')
				else:
					logger.error(
						f'Found value \'{args.env}\' for \'env\' CLI Arg, but not defined in \'envs\' App Config Attr. Exiting.'
					)
					exit(1)
			else:
				if app_config.env in app_config.envs.keys():
					logger.info(
						f'Setting \'env\' Config Attr to \'env\' App Config Attr value \'{app_config.env}\''
					)
					Config.env = app_config.env
				elif app_config.env.startswith('$__env.'):
					logger.info(
						'Found Env Variable for \'env\' App Config Attr. Attempting to process it.'
					)
					env_env_var = app_config.env.replace('$__env.', '')
					env = os.getenv(env_env_var)
					if env:
						logger.info(
							f'Setting \'env\' Config Attr to Env Variable \'{env_env_var}\' value \'{env}\'.'
						)
						Config.env = env
					else:
						logger.error(f'No value found for Env Variable \'{env_env_var}\'. Exiting.')
						exit(1)
				else:
					logger.error(
						f'Found value \'{args.env}\' for \'env\' CLI Arg, but not defined in \'envs\' App Config Attr. Exiting.'
					)
					exit(1)
			logger.info(
				f'Beginning to extract Config Attrs defined in selected \'env\', \'{Config.env}\', to App Config Attrs.'
			)
			for config_attr in dir(app_config.envs[Config.env]):
				if (
					config_attr.startswith('__')
					or getattr(app_config.envs[Config.env], config_attr) == None
				):
					continue
				logger.info(f'Extracting \'{config_attr}\' Config Attr to App Config Attr')
				setattr(
					app_config,
					config_attr,
					getattr(app_config.envs[Config.env], config_attr),
				)
				setattr(
					Config,
					config_attr,
					getattr(app_config.envs[Config.env], config_attr),
				)
		# [DOC] Check port Config Attr
		if not custom_launch and app_config.port:
			if args.port:
				logger.info(
					f'Ignoring \'port\' App Config Attr in favour of \'port\' CLI Arg with value \'{args.port}\'.'
				)
				try:
					Config.port = int(args.port)
				except:
					logger.error(f'Port should be in integer format. Exiting.')
					exit(1)
			else:
				logger.info('Found \'port\' App Config Attr. Attempting to process it.')
				if type(app_config.port) == int:
					Config.port = app_config.port
					logger.info(f'Setting \'port\' Config Attr to \'{Config.port}\'.')
				elif type(app_config.port) == str and app_config.port.startswith('$__env.'):
					logger.info(
						'Found Env Variable for \'port\' App Config Attr. Attempting to process it.'
					)
					port_env_var = app_config.port.replace('$__env.', '')
					port = os.getenv(port_env_var)
					if port:
						logger.info(
							f'Setting \'port\' Config Attr to Env Variable \'{port_env_var}\' value \'{port}\'.'
						)
						try:
							Config.port = int(port)
						except:
							logger.error(
								f'Env Variable \'{port_env_var}\' value \'{port}\' can\'t be converted to integer. Exiting.'
							)
							exit(1)
					else:
						logger.error(f'No value found for Env Variable \'{port_env_var}\'. Exiting.')
						exit(1)
				else:
					logger.error(
						f'Invalid value type for \'port\' Config Attr with value \'{app_config.port}\'. Exiting.'
					)
					exit(1)
		# [DOC] Check debug Config Attr
		if app_config.debug:
			if args.debug:
				logger.info(
					f'Ignoring \'debug\' App Config Attr in favour of \'debug\' CLI Arg with value \'{args.debug}\'.'
				)
				Config.debug = args.debug
			else:
				logger.info('Found \'debug\' App Config Attr. Attempting to process it.')
				if type(app_config.debug) == bool:
					Config.debug = app_config.debug
					logger.info(f'Setting \'debug\' Config Attr to \'{Config.debug}\'.')
				elif type(app_config.debug) == str and app_config.debug.startswith('$__env.'):
					logger.info(
						'Found Env Variable for \'debug\' App Config Attr. Attempting to process it.'
					)
					debug_env_var = app_config.debug.replace('$__env.', '')
					debug = os.getenv(debug_env_var)
					if debug:
						logger.info(
							f'Setting \'debug\' Config Attr to Env Variable \'{debug_env_var}\' as \'True\'.'
						)
						Config.debug = True
					else:
						logger.info(
							f'No value found for Env Variable \'{debug_env_var}\'. Setting \'debug\' to \'False\'.'
						)
						Config.debug = False
				else:
					logger.error(
						f'Invalid value type for \'debug\' Config Attr with value \'{app_config.debug}\'. Exiting.'
					)
					exit(1)
				if Config.debug:
					logger.setLevel(logging.DEBUG)
		# [DOC] Check force_admin_check Config Attr
		if not custom_launch and app_config.force_admin_check:
			if args.force_admin_check:
				logger.info(
					f'Ignoring \'force_admin_check\' App Config Attr in favour of \'force_admin_check\' CLI Arg with value \'{args.force_admin_check}\'.'
				)
				Config.force_admin_check = True
			else:
				logger.info(
					'Found \'force_admin_check\' App Config Attr. Attempting to process it.'
				)
				if type(app_config.force_admin_check) == bool:
					Config.force_admin_check = app_config.force_admin_check
					logger.info(
						f'Setting \'force_admin_check\' Config Attr to \'{Config.force_admin_check}\'.'
					)
				elif type(
					app_config.force_admin_check
				) == str and app_config.force_admin_check.startswith('$__env.'):
					logger.info(
						'Found Env Variable for \'force_admin_check\' App Config Attr. Attempting to process it.'
					)
					check_env_var = app_config.force_admin_check.replace('$__env.', '')
					check = os.getenv(check_env_var)
					if check:
						logger.info(
							f'Setting \'force_admin_check\' Config Attr to Env Variable \'{check_env_var}\' as \'True\'.'
						)
						Config.force_admin_check = True
					else:
						logger.info(
							f'No value found for Env Variable \'{check_env_var}\'. Setting \'force_admin_check\' to \'False\'.'
						)
						Config.force_admin_check = False
				else:
					logger.error(
						f'Invalid value type for \'force_admin_check\' Config Attr with value \'{app_config.force_admin_check}\'. Exiting.'
					)
					exit(1)

		# [DOC] Process other app config attrs as PACKAGE_CONFIG
		_process_config(config=app_config)
	except SystemExit as e:
		exit(e)
	except:
		logger.error(
			'An unexpected exception happened while attempting to process Nawah app. Exception details:'
		)
		logger.error(traceback.format_exc())
		logger.error('Exiting.')
		exit(1)

	if not custom_launch:
		asyncio.run(run_app())
	elif custom_launch == 'generate_ref':

		async def _():
			from nawah.utils import _import_modules, _generate_ref

			await _import_modules()
			_generate_ref()

		asyncio.run(_())
	elif custom_launch == 'generate_models':

		async def _():
			from nawah.utils import _import_modules, _generate_models

			await _import_modules()
			_generate_models()

		asyncio.run(_())



from nawah.config import Config
from nawah.classes import ATTR, EXTN

from typing import Dict, List

import logging, datetime, re, inspect

logger = logging.getLogger('nawah')


def _generate_ref():
	# [DOC] Initialise _api_ref Config Attr
	Config._api_ref = '# API Reference\n'
	# [DOC] Iterate over packages in ascending order
	for package in sorted(Config.modules_packages.keys()):
		# [DOC] Add package header
		Config._api_ref += f'- - -\n## Package: {package.replace("modules.", "")}\n'
		if not len(Config.modules_packages[package]):
			Config._api_ref += f'No modules\n'
		# [DOC] Iterate over package modules in ascending order
		for module in sorted(Config.modules_packages[package]):
			# [DOC] Add module header
			Config._api_ref += f'### Module: {module}\n'
			# [DOC] Add module description
			Config._api_ref += f'{Config.modules[module].__doc__}\n'
			# [DOC] Add module attrs header
			Config._api_ref += '#### Attrs\n'
			# [DOC] Iterate over module attrs to add attrs types, defaults (if any)
			for attr in Config.modules[module].attrs.keys():
				attr_ref = f'* {attr}:\n'
				if Config.modules[module].attrs[attr]._desc:
					attr_ref += f'  * {Config.modules[module].attrs[attr]._desc}\n'
				attr_ref += f'  * Type: `{Config.modules[module].attrs[attr]}`\n'
				for default_attr in Config.modules[module].defaults.keys():
					if (
						default_attr == attr
						or default_attr.startswith(f'{attr}.')
						or default_attr.startswith(f'{attr}:')
					):
						if type(Config.modules[module].defaults[default_attr]) == ATTR:
							attr_ref += f'  * Default [{default_attr}]:\n'
							attr_ref += f'	* Attr Type TYPE: `{Config.modules[module].defaults[default_attr]._args["type"]}`\n'
						else:
							attr_ref += f'  * Default [{default_attr}]: {Config.modules[module].defaults[default_attr]}\n'
				Config._api_ref += attr_ref
			if Config.modules[module].diff:
				Config._api_ref += f'#### Attrs Diff: {Config.modules[module].diff}\n'
			# [DOC] Add module methods
			Config._api_ref += '#### Methods\n'
			for method in Config.modules[module].methods.keys():
				Config._api_ref += f'##### Method: {method}\n'
				Config._api_ref += f'* Permissions Sets:\n'
				for permission in Config.modules[module].methods[method].permissions:
					Config._api_ref += f'  * {permission.privilege}\n'
					# [DOC] Add Query Modifier
					if permission.query_mod:
						Config._api_ref += f'	* Query Modifier:\n'
						if type(permission.query_mod) == dict:
							permission.query_mod = [permission.query_mod]
						for i in range(len(permission.query_mod)):
							Config._api_ref += f'	  * Set {i}:\n'
							# [TODO] Improve nested list sets
							if type(permission.query_mod[i]) != dict:
								Config._api_ref += f'		* List: {permission.query_mod[i]}\n'
								continue
							for attr in permission.query_mod[i].keys():
								if type(permission.query_mod[i][attr]) == ATTR:
									Config._api_ref += f'		* {attr}:\n'
									Config._api_ref += (
										f'		  * Attr Type TYPE: {permission.query_mod[i][attr]._args["type"]}\n'
									)
								else:
									Config._api_ref += f'		* {attr}: {permission.query_mod[i][attr]}\n'
					else:
						Config._api_ref += f'	* Query Modifier: None\n'
					# [DOC] Add Doc Modifier
					if permission.doc_mod:
						Config._api_ref += f'	* Doc Modifier:\n'
						if type(permission.doc_mod) == dict:
							permission.doc_mod = [permission.doc_mod]
						for i in range(len(permission.doc_mod)):
							Config._api_ref += f'	  * Set {i}:\n'
							for attr in permission.doc_mod[i].keys():
								if type(permission.doc_mod[i][attr]) == ATTR:
									Config._api_ref += f'		* {attr}:\n'
									Config._api_ref += (
										f'		  * Attr Type TYPE: `{permission.doc_mod[i][attr]._args["type"]}`\n'
									)
								else:
									Config._api_ref += f'		* {attr}: {permission.doc_mod[i][attr]}\n'
					else:
						Config._api_ref += f'	* Doc Modifier: None\n'
				# [DOC] Add Query Args
				if Config.modules[module].methods[method].query_args:
					Config._api_ref += f'* Query Args Sets:\n'
					for query_args_set in Config.modules[module].methods[method].query_args:
						Config._api_ref += f'  * `{query_args_set}`\n'
				else:
					Config._api_ref += f'* Query Args Sets: None\n'
				# [DOC] Add Doc Args
				if Config.modules[module].methods[method].doc_args:
					Config._api_ref += f'* DOC Args Sets:\n'
					for doc_args_set in Config.modules[module].methods[method].doc_args:
						Config._api_ref += f'  * `{doc_args_set}`\n'
				else:
					Config._api_ref += f'* Doc Args Sets: None\n'
			# [DOC] Add module extns
			if Config.modules[module].extns.keys():
				Config._api_ref += '#### Extended Attrs\n'
				for attr in Config.modules[module].extns.keys():
					Config._api_ref += f'* {attr}:\n'
					if type(Config.modules[module].extns[attr]) == EXTN:
						Config._api_ref += (
							f'  * Module: \'{Config.modules[module].extns[attr].module}\'\n'
						)
						Config._api_ref += (
							f'  * Extend Attrs: \'{Config.modules[module].extns[attr].attrs}\'\n'
						)
						Config._api_ref += f'  * Force: \'{Config.modules[module].extns[attr].force}\'\n'
					elif type(Config.modules[module].extns[attr]) == ATTR:
						Config._api_ref += (
							f'  * Attr Type TYPE: `{Config.modules[module].extns[attr]._args["type"]}`\n'
						)
			else:
				Config._api_ref += '#### Extended Attrs: None\n'
			# [DOC] Add module cache sets
			if Config.modules[module].cache:
				Config._api_ref += '#### Cache Sets\n'
				for i in range(len(Config.modules[module].cache)):
					Config._api_ref += f'* Set {i}:\n'
					Config._api_ref += f'  * CACHE condition: `{_extract_lambda_body(Config.modules[module].cache[i].condition)}`\n'
					Config._api_ref += f'  * CACHE period: {Config.modules[module].cache[i].period}\n'
			else:
				Config._api_ref += '#### Cache Sets: None\n'
			# [DOC] Add module analytics sets
			if Config.modules[module].analytics:
				Config._api_ref += '#### Analytics Sets\n'
				for i in range(len(Config.modules[module].analytics)):
					Config._api_ref += f'* Set {i}:\n'
					Config._api_ref += f'  * ANALYTIC condition: `{_extract_lambda_body(Config.modules[module].analytics[i].condition)}`\n'
					Config._api_ref += f'  * ANALYTIC doc: `{_extract_lambda_body(Config.modules[module].analytics[i].doc)}`\n'
			else:
				Config._api_ref += '#### Analytics Sets: None\n'
	import os

	ref_file = os.path.join(
		Config._app_path,
		'refs',
		f'NAWAH_API_REF_{datetime.datetime.utcnow().strftime("%d-%b-%Y")}.md',
	)
	with open(ref_file, 'w') as f:
		f.write(Config._api_ref)
		logger.info(f'API reference generated and saved to: \'{ref_file}\'. Exiting.')
		exit(0)


def _extract_lambda_body(lambda_func):
	lambda_body = re.sub(
		r'^[a-z]+\s*=\s*lambda\s', '', inspect.getsource(lambda_func).strip()
	)
	if lambda_body.endswith(','):
		lambda_body = lambda_body[:-1]
	return lambda_body



from nawah.config import Config
from nawah.classes import ATTR

from bson import ObjectId
from typing import Any

import logging, random, datetime, re, math

logger = logging.getLogger('nawah')


def generate_attr(*, attr_type: ATTR) -> Any:
	attr_val: Any

	if attr_type._type == 'ANY':
		return '__any'

	elif attr_type._type == 'ACCESS':
		return {'anon': True, 'users': [], 'groups': []}

	elif attr_type._type == 'BOOL':
		attr_val = random.choice([True, False])
		return attr_val

	elif attr_type._type == 'COUNTER':
		counter_groups = re.findall(
			r'(\$__(?:values:[0-9]+|counters\.[a-z0-9_]+))', attr_type._args['pattern']
		)
		attr_val = attr_type._args['pattern']
		for group in counter_groups:
			for group in counter_groups:
				if group.startswith('$__values:'):
					value_callable = attr_type._args['values'][int(group.replace('$__values:', ''))]
					attr_val = attr_val.replace(
						group, str(value_callable(skip_events=[], env={}, query=[], doc={}))
					)
				elif group.startswith('$__counters.'):
					attr_val = attr_val.replace(group, str(42))
		return attr_val

	elif attr_type._type == 'DATE':
		if attr_type._args['ranges']:
			datetime_range = attr_type._args['ranges'][0]
			# [DOC] Be lazy! find a whether start, end of range is a datetime and base the value on it
			if datetime_range[0][0] in ['+', '-'] and datetime_range[1][0] in ['+', '-']:
				# [DOC] Both start, end are dynamic, process start
				datetime_range_delta = {}
				if datetime_range[0][-1] == 'd':
					datetime_range_delta = {'days': int(datetime_range[0][:-1])}
				elif datetime_range[0][-1] == 'w':
					datetime_range_delta = {'weeks': int(datetime_range[0][:-1])}
				attr_val = (
					(datetime.datetime.utcnow() + datetime.timedelta(**datetime_range_delta))
					.isoformat()
					.split('T')[0]
				)
			else:
				if datetime_range[0][0] not in ['+', '-']:
					attr_val = datetime_range[0]
				else:
					attr_val = (
						(datetime.datetime.fromisoformat(datetime_range[1]) - datetime.timedelta(days=1))
						.isoformat()
						.split('T')[0]
					)
		else:
			attr_val = datetime.datetime.utcnow().isoformat().split('T')[0]
		return attr_val

	elif attr_type._type == 'DATETIME':
		if attr_type._args['ranges']:
			datetime_range = attr_type._args['ranges'][0]
			# [DOC] Be lazy! find a whether start, end of range is a datetime and base the value on it
			if datetime_range[0][0] in ['+', '-'] and datetime_range[1][0] in ['+', '-']:
				# [DOC] Both start, end are dynamic, process start
				datetime_range_delta = {}
				if datetime_range[0][-1] == 'd':
					datetime_range_delta = {'days': int(datetime_range[0][:-1])}
				elif datetime_range[0][-1] == 's':
					datetime_range_delta = {'seconds': int(datetime_range[0][:-1])}
				elif datetime_range[0][-1] == 'm':
					datetime_range_delta = {'minutes': int(datetime_range[0][:-1])}
				elif datetime_range[0][-1] == 'h':
					datetime_range_delta = {'hours': int(datetime_range[0][:-1])}
				elif datetime_range[0][-1] == 'w':
					datetime_range_delta = {'weeks': int(datetime_range[0][:-1])}
				attr_val = (
					datetime.datetime.utcnow() + datetime.timedelta(**datetime_range_delta)
				).isoformat()
			else:
				if datetime_range[0][0] not in ['+', '-']:
					attr_val = datetime_range[0]
				else:
					attr_val = (
						datetime.datetime.fromisoformat(datetime_range[1]) - datetime.timedelta(days=1)
					).isoformat()
		else:
			attr_val = datetime.datetime.utcnow().isoformat()
		return attr_val

	elif attr_type._type == 'KV_DICT':
		attr_val = {}
		if attr_type._args['req']:
			attr_val = {
				generate_attr(attr_type=ATTR.LITERAL(literal=[req])): generate_attr(
					attr_type=attr_type._args['val']
				)
				for req in attr_type._args['req']
			}
		for _ in range(attr_type._args['min'] or 0):
			attr_val[generate_attr(attr_type=attr_type._args['key'])] = generate_attr(
				attr_type=attr_type._args['val']
			)
		if len(attr_val.keys()) < (attr_type._args['min'] or 0):
			attr_val = generate_attr(attr_type=attr_type)
		return attr_val

	elif attr_type._type == 'TYPED_DICT':
		attr_val = {
			child_attr: generate_attr(attr_type=attr_type._args['dict'][child_attr])
			for child_attr in attr_type._args['dict'].keys()
		}
		return attr_val

	elif attr_type._type == 'EMAIL':
		attr_val = f'some-{math.ceil(random.random() * 10000)}@mail.provider.com'
		if attr_type._args['allowed_domains']:
			if attr_type._args['strict']:
				domain = 'mail.provider.com'
			else:
				domain = 'provider.com'
			attr_val = attr_val.replace(
				domain, random.choice(attr_type._args['allowed_domains'])
			)
		return attr_val

	elif attr_type._type == 'FILE':
		attr_file_type = 'text/plain'
		attr_file_extension = 'txt'
		if attr_type._args['types']:
			for file_type in attr_type._args['types']:
				if '/' in file_type:
					attr_file_type = file_type
				if '*.' in file_type:
					attr_file_extension = file_type.replace('*.', '')
		file_name = f'__file-{math.ceil(random.random() * 10000)}.{attr_file_extension}'
		return {
			'name': file_name,
			'lastModified': 100000,
			'type': attr_file_type,
			'size': 6,
			'content': b'__file',
		}

	elif attr_type._type == 'FLOAT':
		if attr_type._args['ranges']:
			attr_val = random.choice(
				range(
					math.ceil(attr_type._args['ranges'][0][0]),
					math.floor(attr_type._args['ranges'][0][1]),
				)
			)
			if (
				attr_val != attr_type._args['ranges'][0][0]
				and (attr_val - 0.01) != attr_type._args['ranges'][0][0]
			):
				attr_val -= 0.01
			elif (attr_val + 0.01) < attr_type._args['ranges'][0][1]:
				attr_val += 0.01
			else:
				attr_val = float(attr_val)
		else:
			attr_val = random.random() * 10000
		return attr_val

	elif attr_type._type == 'GEO':
		return {
			'type': 'Point',
			'coordinates': [
				math.ceil(random.random() * 100000) / 1000,
				math.ceil(random.random() * 100000) / 1000,
			],
		}

	elif attr_type._type == 'ID':
		return ObjectId()

	elif attr_type._type == 'INT':
		if attr_type._args['ranges']:
			attr_val = random.choice(
				range(attr_type._args['ranges'][0][0], attr_type._args['ranges'][0][1])
			)
		else:
			attr_val = math.ceil(random.random() * 10000)
		return attr_val

	elif attr_type._type == 'IP':
		return '127.0.0.1'

	elif attr_type._type == 'LIST':
		return [
			generate_attr(attr_type=random.choice(attr_type._args['list']))
			for _ in range(attr_type._args['min'] or 0)
		]

	elif attr_type._type == 'LOCALE':
		return {
			locale: f'__locale-{math.ceil(random.random() * 10000)}' for locale in Config.locales
		}

	elif attr_type._type == 'LOCALES':
		return Config.locale

	elif attr_type._type == 'PHONE':
		attr_phone_code = '000'
		if attr_type._args['codes']:
			attr_phone_code = random.choice(attr_type._args['codes'])
		return f'+{attr_phone_code}{math.ceil(random.random() * 10000)}'

	elif attr_type._type == 'STR':
		if attr_type._args['pattern']:
			logger.warning('Generator for Attr Type STR can\'t handle patterns. Ignoring.')
		return f'__str-{math.ceil(random.random() * 10000)}'

	elif attr_type._type == 'TIME':
		if attr_type._args['ranges']:
			datetime_range = attr_type._args['ranges'][0]
			# [DOC] Be lazy! find a whether start, end of range is a datetime and base the value on it
			if datetime_range[0][0] in ['+', '-'] and datetime_range[1][0] in ['+', '-']:
				# [DOC] Both start, end are dynamic, process start
				datetime_range_delta = {}
				if datetime_range[0][-1] == 's':
					datetime_range_delta = {'seconds': int(datetime_range[0][:-1])}
				elif datetime_range[0][-1] == 'm':
					datetime_range_delta = {'minutes': int(datetime_range[0][:-1])}
				elif datetime_range[0][-1] == 'h':
					datetime_range_delta = {'hours': int(datetime_range[0][:-1])}
				attr_val = (
					(datetime.datetime.utcnow() + datetime.timedelta(**datetime_range_delta))
					.isoformat()
					.split('T')[1]
				)
			else:
				if datetime_range[0][0] not in ['+', '-']:
					attr_val = datetime_range[0]
				else:
					# [REF]: https://stackoverflow.com/a/656394/2393762
					attr_val = (
						(
							datetime.datetime.combine(
								datetime.date.today(), datetime.time.fromisoformat(datetime_range[1])
							)
							- datetime.timedelta(minutes=1)
						)
						.isoformat()
						.split('T')[1]
					)
		else:
			attr_val = datetime.datetime.utcnow().isoformat().split('T')[1]
		return attr_val

	elif attr_type._type == 'URI_WEB':
		attr_val = f'https://sub.domain.com/page-{math.ceil(random.random() * 10000)}/'
		if attr_type._args['allowed_domains']:
			if attr_type._args['strict']:
				domain = 'sub.domain.com'
			else:
				domain = 'domain.com'
			attr_val = attr_val.replace(
				domain, random.choice(attr_type._args['allowed_domains'])
			)
		return attr_val

	elif attr_type._type == 'LITERAL':
		attr_val = random.choice(attr_type._args['literal'])
		return attr_val

	elif attr_type._type == 'UNION':
		attr_val = generate_attr(attr_type=random.choice(attr_type._args['union']))
		return attr_val

	raise Exception(f'Unknown generator attr \'{attr_type}\'')


from nawah.classes import ATTR
from nawah.enums import NAWAH_VALUES

from typing import Dict, Any

import copy


def encode_attr_type(*, attr_type: ATTR) -> Dict[str, Any]:
	encoded_attr_type: Dict[str, Any] = {
		'type': attr_type._type,
		'args': copy.deepcopy(attr_type._args),
		'allow_none': attr_type._default != NAWAH_VALUES.NONE_VALUE,
		'default': attr_type._default
		if attr_type._default != NAWAH_VALUES.NONE_VALUE
		else None,
	}
	# [DOC] Process args of type ATTR
	if attr_type._type == 'LIST':
		for i in range(len(attr_type._args['list'])):
			encoded_attr_type['args']['list'][i] = encode_attr_type(
				attr_type=attr_type._args['list'][i]
			)
	elif attr_type._type == 'TYPED_DICT':
		for dict_attr in attr_type._args['dict'].keys():
			encoded_attr_type['args']['dict'][dict_attr] = encode_attr_type(
				attr_type=attr_type._args['dict'][dict_attr]
			)
	elif attr_type._type == 'KV_DICT':
		encoded_attr_type['args']['key'] = encode_attr_type(attr_type=attr_type._args['key'])
		encoded_attr_type['args']['val'] = encode_attr_type(attr_type=attr_type._args['val'])
	elif attr_type._type == 'UNION':
		for i in range(len(attr_type._args['union'])):
			encoded_attr_type['args']['union'][i] = encode_attr_type(
				attr_type=attr_type._args['union'][i]
			)
	elif attr_type._type == 'TYPE':
		if callable(encoded_attr_type['args']['type']):
			raise Exception('Attr Type TYPE with callable \'type\' can\'t be encoded.')

		del encoded_attr_type['args']['func']

	return encoded_attr_type



from nawah.config import Config

import argparse

from ._launch import launch


def generate_ref(args: argparse.Namespace):
	# [DOC] Update Config with Nawah framework CLI args
	from nawah.config import Config

	Config.generate_ref = True
	launch(args=args, custom_launch='generate_ref')


def generate_models(args: argparse.Namespace):
	# [DOC] Update Config with Nawah framework CLI args
	from nawah.config import Config

	Config.generate_models = True
	launch(args=args, custom_launch='generate_models')



from nawah.config import Config
from nawah.classes import ATTR, DictObj

from typing import Dict, List, TYPE_CHECKING

import logging, datetime, re

if TYPE_CHECKING:
	from nawah.base_module import BaseModule

logger = logging.getLogger('nawah')


def _generate_models():
	# [DOC] Initialise _api_models Config Attr
	Config._api_models = '// Nawah Models\n'
	# [DOC] Add interface for DOC, LOCALE, LOCALES, FILE typing
	Config._api_models += 'interface Doc { _id: string; };\n'
	Config._api_models += 'export interface LOCALE { '
	for locale in Config.locales:
		Config._api_models += locale
		if locale != Config.locale:
			Config._api_models += '?'
		Config._api_models += ': string; '
	Config._api_models += '};\n'
	Config._api_models += (
		'export type LOCALES = \'' + '\' | \''.join(Config.locales) + '\';\n'
	)
	Config._api_models += 'export type ID<T> = string & T;\n'
	Config._api_models += 'export interface FILE<T> { name: string; lastModified: number; type: T; size: number; content: string | boolean; };\n'
	# [DOC] Iterate over packages in ascending order
	for package in sorted(Config.modules_packages.keys()):
		# [DOC] Add package header
		Config._api_models += f'\n// Package: {package.replace("modules.", "")}\n'
		if not len(Config.modules_packages[package]):
			Config._api_models += f'// No modules\n'
		# [DOC] Iterate over package modules in ascending order
		for module in sorted(Config.modules_packages[package]):
			module_class = str(Config.modules[module].__class__).split('.')[-1].split('\'')[0]
			# [DOC] Add module header
			Config._api_models += f'// Module: {module_class}\n'

			# [DOC] Add module interface definition
			Config._api_models += f'export interface {module_class} extends String, Doc {{\n'
			# [DOC] Iterate over module attrs to add attrs types, defaults (if any)
			for attr in Config.modules[module].attrs.keys():
				attr_model = ''
				if Config.modules[module].attrs[attr]._desc:
					attr_model += (
						f'\t// @property {{__TYPE__}} {Config.modules[module].attrs[attr]._desc}\n'
					)
				attr_model += f'\t{attr}__DEFAULT__: __TYPE__;\n'
				for default_attr in Config.modules[module].defaults.keys():
					if (
						default_attr == attr
						or default_attr.startswith(f'{attr}.')
						or default_attr.startswith(f'{attr}:')
					):
						# [DOC] Attr is in defaults, indicate the same
						attr_model = attr_model.replace('__DEFAULT__', '?')
				# [DOC] Attempt to replace __DEFAULT__ with empty string if it still exists, effectively no default value
				attr_model = attr_model.replace('__DEFAULT__', '')

				# [DOC] Add typing
				attr_model = attr_model.replace(
					'__TYPE__',
					_generate_model_typing(
						module=Config.modules[module],
						attr_name=attr,
						attr_type=Config.modules[module].attrs[attr],
					),
				)

				Config._api_models += attr_model

			# [DOC] Add closing braces
			Config._api_models += '};\n'
		Config._api_models += '\n'
	import os

	models_file = os.path.join(
		Config._app_path,
		'models',
		f'NAWAH_API_MODELS_{datetime.datetime.utcnow().strftime("%d-%b-%Y")}.ts',
	)
	with open(models_file, 'w') as f:
		f.write(Config._api_models)
		logger.info(f'API models generated and saved to: \'{models_file}\'. Exiting.')
		exit(0)


def _generate_model_typing(*, module: 'BaseModule', attr_name: str, attr_type: ATTR):
	if attr_type._type == 'ANY':
		return 'any'

	elif attr_type._type == 'ACCESS':
		return '{ anon: boolean; users: Array<string>; groups: Array<string>; }'

	elif attr_type._type == 'BOOL':
		return 'boolean'

	elif attr_type._type == 'COUNTER':
		return 'string'

	elif attr_type._type == 'DATE':
		return 'string'

	elif attr_type._type == 'DATETIME':
		return 'string'

	elif attr_type._type == 'DYNAMIC_ATTR':
		return '{ type: string; args: { [key: string]: any; }; allow_none?: boolean; default: any; }'

	elif attr_type._type == 'DYNAMIC_VAL':
		return 'any'

	elif attr_type._type == 'KV_DICT':
		key_typing = _generate_model_typing(
			module=module, attr_name=attr_name, attr_type=attr_type._args['key']
		)
		val_typing = _generate_model_typing(
			module=module, attr_name=attr_name, attr_type=attr_type._args['val']
		)
		return f'{{ [key: {key_typing}]: {val_typing} }}'

	elif attr_type._type == 'TYPED_DICT':
		typing = '{ '
		for child_attr_type in attr_type._args['dict'].keys():
			typing += child_attr_type
			typing += ': '
			typing += _generate_model_typing(
				module=module,
				attr_name=attr_name,
				attr_type=attr_type._args['dict'][child_attr_type],
			)
			typing += '; '
		typing += '}'
		return typing

	elif attr_type._type == 'EMAIL':
		return 'string'

	elif attr_type._type == 'FILE':
		types = 'string'
		if attr_type._args['types']:
			types = '\'' + '\' | \''.join(attr_type._args['types']) + '\''
			if '*' in types:
				types = 'string'

		return f'FILE<{types}>'

	elif attr_type._type == 'FLOAT':
		return 'number'

	elif attr_type._type == 'GEO':
		return '{ type: \'Point\'; coordinates: [number, number]; }'

	elif attr_type._type == 'ID':
		for attr in module.extns.keys():
			# [TODO] Validate whether Extension Instructions as ATTR breaks this code
			if attr.split('.')[0].split(':')[0] == attr_name:
				# [REF]: https://dev.to/rrampage/snake-case-to-camel-case-and-back-using-regular-expressions-and-python-m9j

				extn_module_class = re.sub(
					r'(.*?)_([a-zA-Z])',
					lambda match: match.group(1) + match.group(2).upper(),
					module.extns[attr].module,
				)
				return f'ID<{extn_module_class[0].upper()}{extn_module_class[1:]}>'
		return 'ID<string>'

	elif attr_type._type == 'INT':
		return 'number'

	elif attr_type._type == 'IP':
		return 'string'

	elif attr_type._type == 'LIST':
		list_typings = []
		for child_attr_type in attr_type._args['list']:
			list_typings.append(
				_generate_model_typing(
					module=module, attr_name=attr_name, attr_type=child_attr_type
				)
			)
		list_typing = ' | '.join(list_typings)
		return f'Array<{list_typing}>'

	elif attr_type._type == 'LOCALE':
		return 'LOCALE'

	elif attr_type._type == 'LOCALES':
		return 'LOCALES'

	elif attr_type._type == 'PHONE':
		return 'string'

	elif attr_type._type == 'STR':
		return 'string'

	elif attr_type._type == 'TIME':
		return 'string'

	elif attr_type._type == 'URI_WEB':
		return 'string'

	elif attr_type._type == 'LITERAL':
		return '\'' + '\' | \''.join(attr_type._args['literal']) + '\''

	elif attr_type._type == 'UNION':
		return ' | '.join(
			[
				_generate_model_typing(
					module=module, attr_name=attr_name, attr_type=child_attr_type
				)
				for child_attr_type in attr_type._args['union']
			]
		)

	elif attr_type._type == 'TYPE':
		return 'any'



from nawah.classes import (
	Query,
	ATTR,
	InvalidCallArgsException,
	InvalidAttrException,
	ConvertAttrException,
)
from nawah.utils import validate_attr

from typing import Union, Dict, Any, List, Literal, Iterable, cast


async def _validate_args(
	*,
	args: Union[Query, Dict[str, Any]],
	args_list_label: str,
	args_list: List[Dict[str, ATTR]],
):
	if not args_list:
		return

	sets_check: List[Dict[str, Literal[True, 'missing', 'invalid', 'convert']]] = []

	for args_set in args_list:
		set_status = True
		set_check = len(sets_check)
		sets_check.append({arg: True for arg in args_set.keys()})

		if args_list_label == 'query':
			args_check: Iterable = args
		elif args_list_label == 'doc':
			args = cast(Dict[str, Any], args)
			args_check = args.keys()

		for arg in args_set.keys():
			if arg not in args_check:
				set_status = False
				sets_check[set_check][arg] = 'missing'
			else:
				try:
					if args_list_label == 'query' and arg[0] != '$':
						for i in range(len(args[arg])):
							if ':' not in arg:
								args[arg][i] = await validate_attr(
									mode='create',
									attr_name=arg,
									attr_type=args_set[arg],
									attr_val=args[arg][i],
								)
							else:
								attr_val = args[arg][i]
								if (arg_oper := arg.split(':')[1]) in ['*', '$eq']:
									if type(args[arg][i]) and arg_oper in args[arg][i].keys():
										attr_val = args[arg][i][arg_oper]
								attr_val = await validate_attr(
									mode='create',
									attr_name=arg,
									attr_type=args_set[arg],
									attr_val=attr_val,
								)
								if arg_oper in ['*', '$eq']:
									args[arg][i] = attr_val
								else:
									args[arg][i] = {arg_oper: attr_val}

					elif args_list_label == 'query' and arg[0] == '$':
						args[arg] = await validate_attr(
							mode='create',
							attr_name=arg,
							attr_type=args_set[arg],
							attr_val=args[arg],
						)
					elif args_list_label == 'doc':
						args[arg] = await validate_attr(
							mode='create',
							attr_name=arg,
							attr_type=args_set[arg],
							attr_val=args[arg],
						)
				except InvalidAttrException:
					set_status = False
					sets_check[set_check][arg] = 'invalid'
				except ConvertAttrException:
					set_status = False
					sets_check[set_check][arg] = 'convert'

		if set_status:
			return

	raise InvalidCallArgsException(sets_check)


from nawah.base_module import BaseModule
from nawah.enums import Event, NAWAH_VALUES
from nawah.classes import ATTR, PERM, EXTN, METHOD
from nawah.config import Config
from nawah.registry import Registry
from nawah.utils import validate_attr, encode_attr_type

from bson import ObjectId

import copy


class User(BaseModule):
	'''`User` module provides data type and controller for users in Nawah eco-system. The permissions of the module methods are designed to be as secure for exposed calls, and as flexible for privileged-access.'''

	collection = 'users'
	attrs = {
		'name': ATTR.LOCALE(desc='Name of the user as `LOCALE`.'),
		'locale': ATTR.LOCALES(desc='Default locale of the user.'),
		'create_time': ATTR.DATETIME(
			desc='Python `datetime` ISO format of the doc creation.'
		),
		'login_time': ATTR.DATETIME(desc='Python `datetime` ISO format of the last login.'),
		'groups': ATTR.LIST(
			desc='List of `_id` for every group the user is member of.',
			list=[ATTR.ID(desc='`_id` of Group doc the user is member of.')],
		),
		'privileges': ATTR.KV_DICT(
			desc='Privileges of the user. These privileges are always available to the user regardless of whether groups user is part of have them or not.',
			key=ATTR.STR(),
			val=ATTR.LIST(list=[ATTR.STR()]),
		),
		'status': ATTR.LITERAL(
			desc='Status of the user to determine whether user has access to the app or not.',
			literal=['active', 'banned', 'deleted', 'disabled_password'],
		),
	}
	defaults = {'login_time': None, 'status': 'active', 'groups': [], 'privileges': {}}
	unique_attrs = []
	methods = {
		'read': METHOD(
			permissions=[
				PERM(privilege='admin'),
				PERM(privilege='read', query_mod={'_id': '$__user'}),
			]
		),
		'create': METHOD(permissions=[PERM(privilege='admin')]),
		'update': METHOD(
			permissions=[
				PERM(privilege='admin', doc_mod={'groups': None}),
				PERM(
					privilege='update',
					query_mod={'_id': '$__user'},
					doc_mod={'groups': None, 'privileges': None},
				),
			],
			query_args={'_id': ATTR.ID()},
		),
		'delete': METHOD(
			permissions=[
				PERM(privilege='admin'),
				PERM(privilege='delete', query_mod={'_id': '$__user'}),
			],
			query_args={'_id': ATTR.ID()},
		),
		'read_privileges': METHOD(
			permissions=[
				PERM(privilege='admin'),
				PERM(privilege='read', query_mod={'_id': '$__user'}),
			],
			query_args={'_id': ATTR.ID()},
		),
		'add_group': METHOD(
			permissions=[PERM(privilege='admin')],
			query_args={'_id': ATTR.ID()},
			doc_args=[{'group': ATTR.ID()}, {'group': ATTR.LIST(list=[ATTR.ID()])}],
		),
		'delete_group': METHOD(
			permissions=[PERM(privilege='admin')],
			query_args={'_id': ATTR.ID(), 'group': ATTR.ID()},
		),
		'retrieve_file': METHOD(permissions=[PERM(privilege='__sys')], get_method=True),
		'create_file': METHOD(permissions=[PERM(privilege='__sys')]),
		'delete_file': METHOD(permissions=[PERM(privilege='__sys')]),
	}

	async def on_read(self, results, skip_events, env, query, doc, payload):
		for i in range(len(results['docs'])):
			user = results['docs'][i]
			for auth_attr in Config.user_attrs.keys():
				del user[f'{auth_attr}_hash']
			if len(Config.user_doc_settings):
				setting_results = await Registry.module('setting').read(
					skip_events=[Event.PERM, Event.ARGS],
					env=env,
					query=[{'user': user._id, 'var': {'$in': Config.user_doc_settings}}],
				)
				user_doc_settings = copy.copy(Config.user_doc_settings)
				if setting_results.args.count:
					for setting_doc in setting_results.args.docs:
						user_doc_settings.remove(setting_doc['var'])
						user[setting_doc['var']] = setting_doc['val']
				# [DOC] Forward-compatibility: If user was created before presence of any user_doc_settings, add them with default value
				for setting_attr in user_doc_settings:
					user[setting_attr] = Config.user_settings[setting_attr].default
					# [DOC] Set NAWAH_VALUES.NONE_VALUE to None if it was default
					if user[setting_attr] == NAWAH_VALUES.NONE_VALUE:
						user[setting_attr] = None
		return (results, skip_events, env, query, doc, payload)

	async def pre_create(self, skip_events, env, query, doc, payload):
		if Event.ARGS not in skip_events:
			doc['groups'] = [ObjectId('f00000000000000000000013')]
		user_settings = {}
		for attr in Config.user_settings.keys():
			if Config.user_settings[attr].type == 'user_sys':
				user_settings[attr] = copy.deepcopy(Config.user_settings[attr].default)
			else:
				if attr in doc.keys():
					try:
						await validate_attr(
							mode='create',
							attr_name=attr,
							attr_type=Config.user_settings[attr].val_type,
							attr_val=doc[attr],
						)
						user_settings[attr] = doc[attr]
					except:
						raise self.exception(
							status=400,
							msg=f'Invalid settings attr \'{attr}\' for \'create\' request on module \'CORE_USER\'',
							args={'code': 'INVALID_ATTR'},
						)

				else:
					if Config.user_settings[attr].default == NAWAH_VALUES.NONE_VALUE:
						raise self.exception(
							status=400,
							msg=f'Missing settings attr \'{attr}\' for \'create\' request on module \'CORE_USER\'',
							args={'code': 'MISSING_ATTR'},
						)

					else:
						user_settings[attr] = copy.deepcopy(Config.user_settings[attr].default)
		payload['user_settings'] = user_settings
		return (skip_events, env, query, doc, payload)

	async def on_create(self, results, skip_events, env, query, doc, payload):
		if 'user_settings' in payload.keys():
			for setting in payload['user_settings'].keys():
				setting_results = await Registry.module('setting').create(
					skip_events=[Event.PERM, Event.ARGS],
					env=env,
					doc={
						'user': results['docs'][0]._id,
						'var': setting,
						'val_type': encode_attr_type(attr_type=Config.user_settings[setting].val_type),
						'val': payload['user_settings'][setting],
						'type': Config.user_settings[setting].type,
					},
				)
				if setting_results.status != 200:
					return setting_results
		return (results, skip_events, env, query, doc, payload)

	async def read_privileges(self, skip_events=[], env={}, query=[], doc={}):
		# [DOC] Confirm _id is valid
		results = await self.read(
			skip_events=[Event.PERM], env=env, query=[{'_id': query['_id'][0]}]
		)
		if not results.args.count:
			raise self.exception(
				status=400, msg='User is invalid.', args={'code': 'INVALID_USER'}
			)
		user = results.args.docs[0]
		for group in user.groups:
			group_results = await Registry.module('group').read(
				skip_events=[Event.PERM], env=env, query=[{'_id': group}]
			)
			group = group_results.args.docs[0]
			for privilege in group.privileges.keys():
				if privilege not in user.privileges.keys():
					user.privileges[privilege] = []
				for i in range(len(group.privileges[privilege])):
					if group.privileges[privilege][i] not in user.privileges[privilege]:
						user.privileges[privilege].append(group.privileges[privilege][i])
		return results

	async def add_group(self, skip_events=[], env={}, query=[], doc={}):
		# [DOC] Check for list group attr
		if type(doc['group']) == list:
			for i in range(0, len(doc['group']) - 1):
				await self.add_group(
					skip_events=skip_events,
					env=env,
					query=query,
					doc={'group': doc['group'][i]},
				)
			doc['group'] = doc['group'][-1]
		# [DOC] Confirm all basic args are provided
		doc['group'] = ObjectId(doc['group'])
		# [DOC] Confirm group is valid
		results = await Registry.module('group').read(
			skip_events=[Event.PERM], env=env, query=[{'_id': doc['group']}]
		)
		if not results.args.count:
			raise self.exception(
				status=400, msg='Group is invalid.', args={'code': 'INVALID_GROUP'}
			)
		# [DOC] Get user details
		results = await self.read(skip_events=[Event.PERM], env=env, query=query)
		if not results.args.count:
			raise self.exception(
				status=400, msg='User is invalid.', args={'code': 'INVALID_USER'}
			)
		user = results.args.docs[0]
		# [DOC] Confirm group was not added before
		if doc['group'] in user.groups:
			raise self.exception(
				status=400,
				msg='User is already a member of the group.',
				args={'code': 'GROUP_ADDED'},
			)
		user.groups.append(doc['group'])
		# [DOC] Update the user
		results = await self.update(
			skip_events=[Event.PERM], env=env, query=query, doc={'groups': user.groups}
		)
		# [DOC] if update fails, return update results
		if results.status != 200:
			return results
		# [DOC] Check if the updated User doc belongs to current session and update it
		if env['session'].user._id == user._id:
			user_results = await self.read_privileges(
				skip_events=[Event.PERM], env=env, query=[{'_id': user._id}]
			)
			env['session']['user'] = user_results.args.docs[0]

		return results

	async def delete_group(self, skip_events=[], env={}, query=[], doc={}):
		# [DOC] Confirm group is valid
		results = await Registry.module('group').read(
			skip_events=[Event.PERM], env=env, query=[{'_id': query['group'][0]}]
		)
		if not results.args.count:
			raise self.exception(
				status=400, msg='Group is invalid.', args={'code': 'INVALID_GROUP'}
			)
		# [DOC] Get user details
		results = await self.read(
			skip_events=[Event.PERM], env=env, query=[{'_id': query['_id'][0]}]
		)
		if not results.args.count:
			raise self.exception(
				status=400, msg='User is invalid.', args={'code': 'INVALID_USER'}
			)
		user = results.args.docs[0]
		# [DOC] Confirm group was not added before
		if query['group'][0] not in user.groups:
			raise self.exception(
				status=400,
				msg='User is not a member of the group.',
				args={'code': 'GROUP_NOT_ADDED'},
			)
		# [DOC] Update the user
		results = await self.update(
			skip_events=[Event.PERM],
			env=env,
			query=[{'_id': query['_id'][0]}],
			doc={'groups': {'$del_val': [query['group'][0]]}},
		)
		# [DOC] if update fails, return update results
		if results.status != 200:
			return results
		# [DOC] Check if the updated User doc belongs to current session and update it
		if env['session'].user._id == user._id:
			user_results = await self.read_privileges(
				skip_events=[Event.PERM], env=env, query=[{'_id': user._id}]
			)
			env['session']['user'] = user_results.args.docs[0]

		return results



from nawah import __version__

from typing import Dict, Any

import argparse, logging, os, sys, re, json, tarfile, tempfile, urllib.request, subprocess, random, string

logger = logging.getLogger('nawah')
__api_level__ = '.'.join(__version__.split('.')[:-1])


def create(args: argparse.Namespace):
	global os, subprocess

	if args.app_name == 'nawah_app':
		logger.error(
			'Value for \'app_name\' CLI Arg is invalid. Name can\'t be \'nawah_app\''
		)
		exit(1)
	elif not re.match(r'^[a-z][a-z0-9_]+$', args.app_name):
		logger.error(
			'Value for \'app_name\' CLI Arg is invalid. Name should have only small letters, numbers, and underscores.'
		)
		exit(1)

	app_path = os.path.realpath(os.path.join(args.app_path, args.app_name))
	progress_path = os.path.realpath(
		os.path.join(args.app_path, args.app_name, 'progress.json')
	)

	progress = None

	if os.path.exists(app_path):
		logger.info(
			'Specified \'app_name\' already existing in \'app_path\'. Attempting to check for earlier progress.'
		)
		if os.path.exists(progress_path):
			logger.info('File \'progress.json\' found. Attempting to process it.')
			try:
				with open(progress_path, 'r') as progress_file:
					progress_config = json.loads(progress_file.read())
					progress = progress_config['step']
					app_config = progress_config['config']
			except Exception as e:
				logger.error(
					'An exception occurred while attempting to process file \'progress.json\'.'
				)
				logger.error(f'Exception details: {e}')
				logger.error('Exiting.')
				exit(1)
		else:
			logger.error('File \'progress.json\' was not found. Exiting.')
			exit(1)

	# [DOC] Populating app_config
	if not progress:
		app_config = _create_step_config(args=args)
		logger.info('This will create an app with the following config:')
		for config_attr, config_set in app_config.items():
			logger.info(f'- {config_attr}: \'{config_set[1]}\'')
	else:
		logger.info('Continuing to create app with loaded progress config:')
		for config_attr, config_set in app_config.items():
			logger.info(f'- {config_attr}: \'{config_set[1]}\'')

	# [DOC] Create app workspace
	if not progress:
		cli_dir, cli_filename = os.path.split(__file__)
		template_archive = os.path.join(cli_dir, 'template.tar.gz')

		logger.info(f'Attempting to extract template archive to: {app_path}')
		with tarfile.open(name=template_archive, mode='r:gz') as archive:
			archive.extractall(path=app_path)
		logger.info('Template archive extracted successfully!')

	if not progress or progress == 1:
		progress = None
		logger.info('Attempting to config app template for new Nawah app.')
		try:
			with open(
				os.path.realpath(os.path.join(args.app_path, args.app_name, 'nawah_app.py')), 'r'
			) as f:
				nawah_app_file = f.read()
			with open(
				os.path.realpath(os.path.join(args.app_path, args.app_name, 'nawah_app.py')), 'w'
			) as f:
				nawah_app_file = nawah_app_file.replace('__PROJECT_NAME__', args.app_name, 2)
				for config_set in app_config.values():
					nawah_app_file = nawah_app_file.replace(config_set[0], config_set[1], 1)
				f.write(nawah_app_file)
		except Exception as e:
			_dump_progress(args=args, app_config=app_config, progress_path=progress_path, step=3)
			logger.error('An exception occurred.')
			logger.error(f'Exception details: {e}')
			logger.error('Exiting.')
			exit(1)

	if not progress or progress == 2:
		progress = None
		try:
			gitignore_path = os.path.realpath(
				os.path.join(args.app_path, args.app_name, '.gitignore')
			)
			with open(gitignore_path, 'r') as f:
				gitignore_file = f.read()
			with open(gitignore_path, 'w') as f:
				f.write(gitignore_file.replace('PROJECT_NAME', args.app_name, 1))
		except Exception as e:
			_dump_progress(args=args, app_config=app_config, progress_path=progress_path, step=4)
			logger.error('An exception occurred.')
			logger.error(f'Exception details: {e}')
			logger.error('Exiting.')
			exit(1)

	if not progress or progress == 3:
		progress = None
		try:
			readme_path = os.path.realpath(
				os.path.join(args.app_path, args.app_name, 'README.md')
			)
			with open(readme_path, 'w') as f:
				f.write(
					f'''# {args.app_name}
This Nawas app project was created with Nawah CLI v{__version__}, with API Level {__api_level__}.'''
				)
		except Exception as e:
			_dump_progress(args=args, app_config=app_config, progress_path=progress_path, step=6)
			logger.error('An exception occurred.')
			logger.error(f'Exception details: {e}')
			logger.error('Exiting.')
			exit(1)

	if not progress or progress == 4:
		progress = None
		try:
			os.rename(
				os.path.realpath(
					os.path.join(args.app_path, args.app_name, 'packages', 'PROJECT_NAME')
				),
				os.path.realpath(
					os.path.join(args.app_path, args.app_name, 'packages', args.app_name)
				),
			)
		except Exception as e:
			_dump_progress(args=args, app_config=app_config, progress_path=progress_path, step=7)
			logger.error('An exception occurred.')
			logger.error(f'Exception details: {e}')
			logger.error('Exiting.')
			exit(1)

	if not progress or progress == 5:
		progress = None
		try:
			test_integration_path = os.path.realpath(
				os.path.join(
					args.app_path, args.app_name, 'tests', 'integration', 'test_integration.py'
				)
			)
			with open(test_integration_path, 'r') as f:
				test_integration_file = f.read()
			with open(test_integration_path, 'w') as f:
				f.write(test_integration_file.replace('PROJECT_NAME', args.app_name.upper(), 1))
		except Exception as e:
			_dump_progress(args=args, app_config=app_config, progress_path=progress_path, step=4)
			logger.error('An exception occurred.')
			logger.error(f'Exception details: {e}')
			logger.error('Exiting.')
			exit(1)

	if not progress or progress == 6:
		progress = None
		try:
			for unit_test_filename in ['test_read.py', 'test_create.py']:
				test_unit_path = os.path.realpath(
					os.path.join(
						args.app_path, args.app_name, 'tests', 'unit', 'doc', unit_test_filename
					)
				)
				with open(test_unit_path, 'r') as f:
					test_unit_file = f.read()
				with open(test_unit_path, 'w') as f:
					f.write(test_unit_file.replace('PROJECT_NAME', args.app_name, 1))
		except Exception as e:
			_dump_progress(args=args, app_config=app_config, progress_path=progress_path, step=4)
			logger.error('An exception occurred.')
			logger.error(f'Exception details: {e}')
			logger.error('Exiting.')
			exit(1)

	logger.info(f'Congrats! Your Nawah app {args.app_name} is successfully created!')


def _create_step_config(*, args: argparse.Namespace):
	app_config = {
		# [REF]: https://stackoverflow.com/a/47073723/2393762
		'admin_password': [
			'__ADMIN_PASSWORD__',
			''.join(
				[
					random.choice(string.ascii_letters + string.digits + string.punctuation)
					for n in range(18)
				]
			)
			.replace('\'', '^')
			.replace('\\', '/'),
		],
		'anon_token_suffix': [
			'__ANON_TOKEN_SUFFIX__',
			''.join([random.choice(string.digits) for n in range(24)]),
		],
	}

	envs_defaults = {
		'dev_local': 'mongodb://localhost',
		'dev_server': 'mongodb://admin:admin@mongodb',
		'prod': 'mongodb://admin:admin@prod',
	}

	if not args.default_config:
		# [DOC] Allow user to specify custom config per the following Config Attrs
		logger.info(
			'Not detected \'default-config\' CLI Arg. Attempting to create Nawah app with custom Config.'
		)
		logger.info(
			'If you would like to have app created with default config, stop the process, and re-run Nawah CLI with \'default-config\' CLI Arg.'
		)
		# [DOC] envs: data_server
		logger.info(
			'\'envs\' Config Attr provides environment-specific configuration. You will be required to specify \'data_server\' for each of the default available Nawah environments, namely \'dev_local\', \'dev_server\', and \'prod\'.'
		)
		for env in envs_defaults.keys():
			while True:
				config_attr_val = input(
					f'\n> What would be the value for \'data_server\' Config Attr for environment \'{env}\'; The connection string to connect to your MongoDB host? [{envs_defaults[env]}]\n- '
				)
				if not config_attr_val:
					logger.info(
						f'Setting \'data_server\' Config Attr for environment \'{env}\' to default \'{envs_defaults[env]}\'.'
					)
					app_config[f'{env}:data_server'] = [
						f'__{env.upper()}_DATA_SERVER__',
						(config_attr_val := envs_defaults[env]),
					]
				else:
					logger.info(
						f'Setting \'data_server\' Config Attr for environment \'{env}\' to: \'{config_attr_val}\'.'
					)
					app_config[f'{env}:data_server'] = [
						f'__{env.upper()}_DATA_SERVER__',
						config_attr_val,
					]
				break
		# [DOC] env
		while True:
			config_attr_val = input(
				'\n> What would be the value for \'env\'; Config Attr for default environnement to be used when invoking Nawah CLI \'launch\' command? [$__env.ENV]\n- '
			)
			if not config_attr_val:
				logger.info('Setting \'env\' Config Attr to default: \'$__env.ENV\'.')
				app_config['data_name'] = ['__ENV__', (config_attr_val := '$__env.ENV')]
				break
			elif config_attr_val not in list(envs_defaults.keys()) and not re.match(
				r'^\$__env\.[A-Za-z_]+$', config_attr_val
			):
				logger.error(
					'\'env\' Config Attr can only be one of the environments names defined in \'envs\' Config Attr, or a valid Env Variable.'
				)
			else:
				logger.info(f'Setting \'env\' Config Attr to: \'{config_attr_val}\'.')
				app_config['env'] = ['__ENV__', config_attr_val]
				break
		# [DOC] data_name
		while True:
			config_attr_val = input(
				'\n> What would be the value for \'data_name\'; Config Attr for database name to be created on \'data_server\'? [nawah_data]\n- '
			)
			if not config_attr_val:
				logger.info('Setting \'data_name\' Config Attr to default: \'nawah_data\'.')
				app_config['data_name'] = ['__DATA_NAME__', (config_attr_val := 'nawah_data')]
				break
			elif not re.match(r'^[a-zA-Z0-9\-_]+$', config_attr_val):
				logger.error(
					'\'data_name\' Config Attr can\'t have special characters other than underscores and hyphens in it.'
				)
			else:
				logger.info(f'Setting \'data_name\' Config Attr to: \'{config_attr_val}\'.')
				app_config['data_name'] = ['__DATA_NAME__', config_attr_val]
				break
		# [DOC] locales
		while True:
			config_attr_val = input(
				'\n> What would be the comma-separated, language_COUNTRY-formatted value for \'locales\'; Config Attr for localisations of your app? [ar_AE, en_AE]\n- '
			)
			if not config_attr_val:
				logger.info('Setting \'locales\' Config Attr to default: \'ar_AE, en_AE\'.')
				locales = ['ar_AE', 'en_AE']
				app_config['locales'] = ['__LOCALES__', (config_attr_val := 'ar_AE\', \'en_AE')]
				break
			else:
				try:
					locales = [
						re.match(r'^([a-z]{2}_[A-Z]{2})$', locale.strip()).group(0)  # type: ignore
						for locale in config_attr_val.split(',')
					]
					logger.info(f'Setting \'locales\' Config Attr to: \'{config_attr_val}\'.')
					app_config['locales'] = ['__LOCALES__', '\', \''.join(locales)]
					break
				except:
					logger.error('An exception occurred while attempting to process value provided.')
					logger.error(
						'\'locales\' Config Attr value should be provided as comma-separated, language_COUNTRY-formatted list of localisations.'
					)
		# [DOC] locale
		while True:
			config_attr_val = input(
				'\n> What would be the value for \'locale\'; Config Attr for default localisation of your app? [first value of \'locales\' Config Attr]\n- '
			)
			if not config_attr_val:
				logger.info(
					f'Setting \'locale\' Config Attr to first value of \'locales\' Config Attr: \'{locales[0]}\'.'
				)
				app_config['locale'] = ['__LOCALE__', (config_attr_val := locales[0])]
				break
			elif config_attr_val not in locales:
				logger.error(
					'\'locale\' Config Attr can only be one of the localisations defined in \'locales\' Config Attr.'
				)
			else:
				logger.info(f'Setting \'locale\' Config Attr to: \'{config_attr_val}\'.')
				app_config['locale'] = ['__LOCALE__', config_attr_val]
				break
		# [DOC] admin_doc: email
		while True:
			config_attr_val = input(
				'\n> What would be the value for \'admin_doc\'.\'email\'; Config Attr? [admin@app.nawah.localhost]\n- '
			)
			if not config_attr_val:
				logger.info(
					'Setting \'admin_doc\'.\'email\' Config Attr to default: \'admin@app.nawah.localhost\'.'
				)
				app_config['admin_doc:email'] = [
					'__ADMIN_DOC_EMAIL__',
					(config_attr_val := 'admin@app.nawah.localhost'),
				]
				break
			elif not re.match(r'^[^@]+@[^@]+\.[^@]+$', config_attr_val):
				logger.error(
					'\'admin_doc\'.\'email\' Config Attr value is not a valid email address.'
				)
			else:
				logger.info(
					f'Setting \'admin_doc\'.\'email\' Config Attr to: \'{config_attr_val}\'.'
				)
				app_config['admin_doc:email'] = ['__ADMIN_DOC_EMAIL__', config_attr_val]
				break
	else:
		for env in envs_defaults.keys():
			app_config[f'{env}:data_server'] = [
				f'__{env.upper()}_DATA_SERVER__',
				envs_defaults[env],
			]
		app_config.update(
			{
				'env': ['__ENV__', '$__env.ENV'],
				'data_name': ['__DATA_NAME__', 'nawah_data'],
				'locales': ['__LOCALES__', '\', \''.join(['ar_AE', 'en_AE'])],
				'locale': ['__LOCALE__', 'ar_AE'],
				'admin_doc:email': ['__ADMIN_DOC_EMAIL__', 'admin@app.nawah.localhost'],
			}
		)

	return app_config


def _dump_progress(
	*, args: argparse.Namespace, app_config: Dict[Any, Any], progress_path: str, step: int
):
	with open(progress_path, 'w') as progress_file:
		progress_file.write(
			json.dumps(
				{
					'step': step,
					'args': {
						'app_path': args.app_path,
						'app_name': args.app_name,
					},
					'config': app_config,
				}
			)
		)



from enum import Enum, auto


class Event(Enum):
	ARGS = auto()
	VALIDATE = auto()
	PERM = auto()
	PRE = auto()
	ON = auto()
	EXTN = auto()
	SOFT = auto()
	DIFF = auto()
	SYS_DOCS = auto()


class DELETE_STRATEGY(Enum):
	SOFT_SKIP_SYS = auto()
	SOFT_SYS = auto()
	FORCE_SKIP_SYS = auto()
	FORCE_SYS = auto()


class LOCALE_STRATEGY(Enum):
	DUPLICATE = auto()
	NONE_VALUE = auto()


class CACHE_STRATEGY(Enum):
	RECREATE = auto()
	DISMISS = auto()


class NAWAH_VALUES(Enum):
	NONE_VALUE = auto()
	ALLOW_MOD = auto()



from bson import ObjectId
from dataclasses import dataclass
from typing import (
	Optional,
	Dict,
	Any,
	TypedDict,
	Literal,
	List,
	Callable,
	Union,
	Optional,
	cast,
	Protocol,
	TYPE_CHECKING,
)

from croniter import croniter

from ._types import NAWAH_DOC

if TYPE_CHECKING:
	from ._attr import ATTR
	from ._types import NAWAH_ENV


CLIENT_APP = TypedDict(
	'CLIENT_APP',
	{
		'name': str,
		'type': Literal['web', 'ios', 'android'],
		'origin': List[str],
		'hash': str,
	},
)


class USER_SETTING:
	type: Literal['user', 'user_sys']
	val_type: 'ATTR'
	default: Optional[Any]

	def __init__(
		self,
		*,
		type: Literal['user', 'user_sys'],
		val_type: 'ATTR',
		default: Optional[Any] = None,
	):
		self.type = type
		self.val_type = val_type
		self.default = default

	def _validate(self):
		from ._attr import ATTR

		# [DOC] Validate type
		if type(self.type) != str or self.type not in ['user', 'user_sys']:
			raise Exception(
				f'Invalid \'type\' of type \'{type(self.type)}\' with required \'user\', or \'user_sys\'.'
			)

		# [DOC] Validate val_type
		if type(self.val_type) != ATTR:
			raise Exception(
				f'Invalid \'val_type\' of type \'{type(self.val_type)}\' with required type \'ATTR\'.'
			)

		# [DOC] Validate val_type Attr Type
		ATTR.validate_type(attr_type=self.val_type)


ANALYTICS_EVENTS = TypedDict(
	'ANALYTICS_EVENTS',
	{
		'app_conn_verified': bool,
		'session_conn_auth': bool,
		'session_user_auth': bool,
		'session_conn_reauth': bool,
		'session_user_reauth': bool,
		'session_conn_deauth': bool,
		'session_user_deauth': bool,
	},
)


class SYS_DOC:
	module: str
	key: Optional[str]
	skip_args: bool
	doc: Optional[NAWAH_DOC]

	@property
	def key_value(self) -> Any:
		if not (self.doc):
			raise Exception(f'SYS_DOC instance was initialised with no \'doc\'.')
		self.key = cast(str, self.key)
		return self.doc[self.key]

	def __init__(
		self,
		*,
		module: str,
		key: str = None,
		skip_args: bool = False,
		doc: NAWAH_DOC = None,
	):
		if doc != None:
			if type(doc) != dict:
				raise Exception(
					f'Argument \'doc\' is not a valid \'NAWAH_DOC\'. Expecting type \'dict\' but got \'{type(dict)}\'.'
				)
			doc = cast(NAWAH_DOC, doc)
			if not key:
				key = '_id'
			if key not in doc.keys():
				raise Exception(f'Attr \'{key}\' is not present on \'doc\'.')
			if key == '_id' and type(doc["_id"]) != ObjectId:
				raise Exception(
					f'Invalid attr \'_id\' of type \'{type(doc["_id"])}\' with required type \'ID\''
				)
		else:
			if key or skip_args:
				raise Exception('Arguments \'attr, skip_args\' should only be used with \'doc\'.')

		self.module = module
		self.key = key
		self.skip_args = skip_args
		self.doc = doc


class JOB_CALLABLE(Protocol):
	def __call__(
		self,
		env: 'NAWAH_ENV',
	) -> bool:
		...


class JOB:
	job: JOB_CALLABLE
	schedule: str
	prevent_disable: bool
	_cron_schedule: croniter
	_next_time: Optional[str] = None
	_disabled: bool = False

	def __init__(
		self,
		*,
		job: JOB_CALLABLE,
		schedule: str,
		prevent_disable: bool = False,
	):
		self.job = job
		self.schedule = schedule
		self.prevent_disable = prevent_disable


class L10N(dict):
	pass


@dataclass
class PACKAGE_CONFIG:
	api_level: Optional[str] = None
	version: Optional[str] = None
	emulate_test: Optional[bool] = None
	debug: Optional[bool] = None
	port: Optional[int] = None
	env: Optional[str] = None
	force_admin_check: Optional[bool] = None
	vars_types: Optional[Dict[str, 'ATTR']] = None
	vars: Optional[Dict[str, Any]] = None
	client_apps: Optional[Dict[str, CLIENT_APP]] = None
	analytics_events: Optional[ANALYTICS_EVENTS] = None
	conn_timeout: Optional[int] = None
	quota_anon_min: Optional[int] = None
	quota_auth_min: Optional[int] = None
	quota_ip_min: Optional[int] = None
	data_server: Optional[str] = None
	data_name: Optional[str] = None
	data_ssl: Optional[bool] = None
	data_ca_name: Optional[str] = None
	data_ca: Optional[str] = None
	data_disk_use: Optional[bool] = None
	data_azure_mongo: Optional[bool] = None
	locales: Optional[List[str]] = None
	locale: Optional[str] = None
	admin_doc: Optional[NAWAH_DOC] = None
	admin_password: Optional[str] = None
	anon_token: Optional[str] = None
	anon_privileges: Optional[Dict[str, List[str]]] = None
	user_attrs: Optional[Dict[str, 'ATTR']] = None
	user_settings: Optional[Dict[str, USER_SETTING]] = None
	user_doc_settings: Optional[List[str]] = None
	groups: Optional[List[Dict[str, Any]]] = None
	default_privileges: Optional[Dict[str, List[str]]] = None
	data_indexes: Optional[List[Dict[str, Any]]] = None
	docs: Optional[List[SYS_DOC]] = None
	jobs: Optional[Dict[str, JOB]] = None
	gateways: Optional[Dict[str, Callable]] = None
	types: Optional[Dict[str, Callable]] = None


@dataclass
class APP_CONFIG(PACKAGE_CONFIG):
	name: Optional[str] = None
	version: Optional[str] = None
	default_package: Optional[str] = None
	debug: Optional[bool] = False
	port: Optional[int] = None
	env: Optional[str] = None
	envs: Optional[Dict[str, PACKAGE_CONFIG]] = None
	force_admin_check: Optional[bool] = None



from nawah import __version__

from distutils.command.install import install

import setuptools


class version(install):
	def run(self):
		print(__version__)


class api_level(install):
	def run(self):
		print('.'.join(__version__.split('.')[:2]))


with open('README.md', 'r') as f:
	long_description = f.read()

with open('./requirements.txt', 'r') as f:
	requirements = f.readlines()

with open('./dev_requirements.txt', 'r') as f:
	dev_requirements = f.readlines()

setuptools.setup(
	name='nawah',
	version=__version__,
	author='Mahmoud Abduljawad',
	author_email='mahmoud@masaar.com',
	description='Nawah framework--Rapid app development framework',
	long_description=long_description,
	long_description_content_type='text/markdown',
	url='https://github.com/nawah-io/nawah_framework',
	package_data={
		'nawah': ['py.typed'],
		'nawah.cli': ['template.tar.gz', 'py.typed'],
		'nawah.base_method': ['py.typed'],
		'nawah.base_module': ['py.typed'],
		'nawah.classes': ['py.typed'],
		'nawah.config': ['py.typed'],
		'nawah.data': ['py.typed'],
		'nawah.enums': ['py.typed'],
		'nawah.gateway': ['py.typed'],
		'nawah.registry': ['py.typed'],
		'nawah.testing': ['py.typed'],
		'nawah.utils': ['py.typed'],
		'nawah.packages': ['py.typed'],
		'nawah.packages.core': ['py.typed'],
	},
	packages=[
		'nawah',
		'nawah.cli',
		'nawah.base_method',
		'nawah.base_module',
		'nawah.classes',
		'nawah.config',
		'nawah.data',
		'nawah.enums',
		'nawah.gateway',
		'nawah.registry',
		'nawah.testing',
		'nawah.utils',
		'nawah.packages',
		'nawah.packages.core',
	],
	project_urls={
		'Docs: Github': 'https://github.com/nawah-io/nawah_docs',
		'GitHub: issues': 'https://github.com/nawah-io/nawah_framework/issues',
		'GitHub: repo': 'https://github.com/nawah-io/nawah_framework',
	},
	classifiers=[
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.8',
		'Development Status :: 5 - Production/Stable',
		'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
		'Operating System :: OS Independent',
		'Topic :: Internet :: WWW/HTTP',
		'Framework :: AsyncIO',
	],
	python_requires='>=3.8',
	install_requires=requirements,
	extras_require={'dev': dev_requirements},
	cmdclass={
		'version': version,
		'api_level': api_level,
	},
	entry_points={
		'console_scripts': {
			'nawah = nawah.__main__:main',
		}
	},
	zip_safe=False,
)



from bson import ObjectId

import json, datetime

from ._base_model import BaseModel
from ._dictobj import DictObj


class JSONEncoder(json.JSONEncoder):
	def default(self, o):
		if isinstance(o, ObjectId):
			return str(o)
		elif isinstance(o, BaseModel) or isinstance(o, DictObj):
			return o._attrs()
		elif type(o) == datetime.datetime:
			return o.isoformat()
		elif type(o) == bytes:
			return True
		try:
			return json.JSONEncoder.default(self, o)
		except TypeError:
			return str(o)


from nawah.config import Config
from nawah.enums import Event, DELETE_STRATEGY, CACHE_STRATEGY, NAWAH_VALUES
from nawah import data as Data
from nawah.utils import (
	validate_doc,
	_expand_attr,
	_update_attr_values,
	_set_none_value,
)
from nawah.classes import (
	DictObj,
	BaseModel,
	Query,
	NAWAH_EVENTS,
	NAWAH_ENV,
	Query,
	NAWAH_QUERY,
	NAWAH_DOC,
	ATTR,
	PERM,
	EXTN,
	METHOD,
	CACHE,
	CACHED_QUERY,
	ANALYTIC,
	PRE_HANDLER_RETURN,
	ON_HANDLER_RETURN,
	MethodException,
	InvalidAttrTypeException,
	InvalidAttrTypeArgException,
	InvalidAttrException,
	MissingAttrException,
	ConvertAttrException,
)
from nawah.base_method import BaseMethod

from typing import (
	List,
	Dict,
	Union,
	Tuple,
	Callable,
	Any,
	TypedDict,
	cast,
	Literal,
	Optional,
	AsyncGenerator,
)

from PIL import Image
from bson import ObjectId
import traceback, logging, datetime, re, sys, io, copy, asyncio

logger = logging.getLogger('nawah')


class BaseModule:
	_nawah_module: bool = True

	collection: Optional[str]
	attrs: Dict[str, ATTR]
	diff: Union[bool, ATTR]
	create_draft: Union[bool, ATTR]
	update_draft: Union[bool, ATTR]
	defaults: Dict[str, Any]
	unique_attrs: List[Union[str, Tuple[str, ...]]]
	extns: Dict[str, Union[EXTN, ATTR]]
	privileges: List[str]
	methods: Dict[str, METHOD]
	cache: List[CACHE]
	analytics: List[ANALYTIC]

	package_name: str
	module_name: str

	def __init__(self):
		if not getattr(self, 'collection', None):
			self.collection = None
		if not getattr(self, 'attrs', None):
			self.attrs = {}
		if not getattr(self, 'diff', None):
			self.diff = False
		if not getattr(self, 'create_draft', None):
			self.create_draft = False
		if not getattr(self, 'update_draft', None):
			self.update_draft = False
		if not getattr(self, 'defaults', None):
			self.defaults = {}
		if not getattr(self, 'unique_attrs', None):
			self.unique_attrs = []
		if not getattr(self, 'extns', None):
			self.extns = {}
		if not getattr(self, 'privileges', None):
			self.privileges = ['read', 'create', 'update', 'delete', 'admin']
		if not getattr(self, 'methods', None):
			self.methods = {}
		if not getattr(self, 'cache', None):
			self.cache = []
		if not getattr(self, 'analytics', None):
			self.analytics = []

		# [DOC] Populate package and module names for in-context use.
		self.package_name = self.__module__.replace('modules.', '').upper().split('.')[-2]
		self.module_name = re.sub(
			r'([A-Z])',
			r'_\1',
			self.__class__.__name__[0].lower() + self.__class__.__name__[1:],
		).lower()

	def _pre_initialise(self) -> None:
		pass

	def _initialise(self) -> None:
		# [DOC] Call _pre_initialise for advanced module initialisation
		self._pre_initialise()
		# [DOC] Check attrs for any invalid type
		for attr in self.attrs.keys():
			try:
				logger.debug(
					f'Attempting to validate Attr Type for \'{attr}\' of module \'{self.module_name}\'.'
				)
				ATTR.validate_type(attr_type=self.attrs[attr])
			except InvalidAttrTypeException as e:
				logger.error(
					f'Invalid Attr Type for \'{attr}\' of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
				)
				exit(1)
			except InvalidAttrTypeArgException as e:
				logger.error(
					f'Invalid Attr Type Arg for \'{attr}\' of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
				)
				exit(1)
			# [DOC] Check default for invalid types, update default value
			for default in self.defaults.keys():
				if (
					default == attr or default.startswith(f'{attr}.') or default.startswith(f'{attr}:')
				):
					if type(self.defaults[default]) == ATTR:
						if self.defaults[default]._type != 'TYPE':
							logger.error(
								f'Invalid Attr Type for default \'{default}\' of module \'{self.module_name}\'. Only Attr Type TYPE is allowed. Exiting.'
							)
							exit(1)
						logger.debug(
							f'Attempting to validate Attr Type of default \'{default}\' of module \'{self.module_name}\'.'
						)
						try:
							ATTR.validate_type(attr_type=self.defaults[default])
						except InvalidAttrTypeException as e:
							logger.error(
								f'Invalid Attr Type for default \'{default}\' of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
							)
							exit(1)
						except InvalidAttrTypeArgException as e:
							logger.error(
								f'Invalid Attr Type Arg for default \'{default}\' of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
							)
							exit(1)
					logger.debug(
						f'Updating default value for attr \'{attr}\' to: \'{self.defaults[default]}\''
					)
					_update_attr_values(
						attr=ATTR.TYPED_DICT(dict=self.attrs),
						value='default',
						value_path=default,
						value_val=self.defaults[default],
					)
			# [DOC] Update extn value
			for extn in self.extns.keys():
				if extn == attr or extn.startswith(f'{attr}.') or extn.startswith(f'{attr}:'):
					logger.debug(f'Updating extn value for attr \'{extn}\' to: \'{self.extns[extn]}\'')
					_update_attr_values(
						attr=ATTR.TYPED_DICT(dict=self.attrs),
						value='extn',
						value_path=extn,
						value_val=self.extns[extn],
					)
		# [DOC] Abstract methods as BaseMethod objects
		for method_name in self.methods.keys():
			method = self.methods[method_name]
			# [DOC] Check value type
			if type(method) != METHOD:
				logger.error(
					f'Invalid method \'{method}\' of module \'{self.module_name}\'. Exiting.'
				)
				exit(1)
			# [DOC] Set Method._module value to create back-link from child to parent
			method._module = self
			method._method_name = method_name
			# [DOC] Check valid method (and initialise it)
			try:
				method._validate()
			except Exception as e:
				logger.error(e)
				logger.error('Exiting.')
				exit(1)
		# [DOC] Check extns for invalid extended attrs
		for attr in self.extns.keys():
			if type(self.extns[attr]) not in [EXTN, ATTR]:
				logger.error(
					f'Invalid extns attr \'{attr}\' of module \'{self.module_name}\'. Exiting.'
				)
				exit(1)
			if type(self.extns[attr]) == ATTR:
				self.extns[attr] = cast(ATTR, self.extns[attr])
				if self.extns[attr]._type != 'TYPE':
					logger.error(
						f'Invalid Attr Type for extn \'{attr}\' of module \'{self.module_name}\'. Only Attr Type TYPE is allowed. Exiting.'
					)
					exit(1)
				logger.debug(
					f'Attempting to validate Attr Type of extn \'{attr}\' of module \'{self.module_name}\'.'
				)
				try:
					ATTR.validate_type(attr_type=self.extns[attr])
				except InvalidAttrTypeException as e:
					logger.error(
						f'Invalid Attr Type for extn \'{attr}\' of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
					)
					exit(1)
				except InvalidAttrTypeArgException as e:
					logger.error(
						f'Invalid Attr Type Arg for extn \'{attr}\' of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
					)
					exit(1)

		# [DOC] Check valid type, value for diff
		if type(self.diff) not in [bool, ATTR]:
			logger.error(f'Invalid diff for module \'{self.module_name}\'. Exiting.')
			exit(1)
		if type(self.diff) == ATTR:
			self.diff = cast(ATTR, self.diff)
			if self.diff._type != 'TYPE':
				logger.error(
					f'Invalid Attr Type for diff of module \'{self.module_name}\'. Only Attr Type TYPE is allowed. Exiting.'
				)
				exit(1)
			logger.debug(
				f'Attempting to validate Attr Type diff of module \'{self.module_name}\'.'
			)
			try:
				ATTR.validate_type(attr_type=self.diff)
			except InvalidAttrTypeException as e:
				logger.error(
					f'Invalid Attr Type for diff of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
				)
				exit(1)
			except InvalidAttrTypeArgException as e:
				logger.error(
					f'Invalid Attr Type Arg for diff of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
				)
				exit(1)

		# [DOC] Check valid types, values for create_draft, update_draft
		for attr in ['create_draft', 'update_draft']:
			if type(getattr(self, attr)) not in [bool, ATTR]:
				logger.error(f'Invalid {attr} for module \'{self.module_name}\'. Exiting.')
				exit(1)
			if type(getattr(self, attr)) == ATTR:
				if getattr(self, attr)._type != 'TYPE':
					logger.error(
						f'Invalid Attr Type for {attr} of module \'{self.module_name}\'. Only Attr Type TYPE is allowed. Exiting.'
					)
					exit(1)
				logger.debug(
					f'Attempting to validate Attr Type {attr} of module \'{self.module_name}\'.'
				)
				try:
					ATTR.validate_type(attr_type=getattr(self, attr))
				except InvalidAttrTypeException as e:
					logger.error(
						f'Invalid Attr Type for {attr} of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
					)
					exit(1)
				except InvalidAttrTypeArgException as e:
					logger.error(
						f'Invalid Attr Type Arg for {attr} of module \'{self.module_name}\'. Original validation error: {str(e)}. Exiting.'
					)
					exit(1)

		logger.debug(f'Initialised module {self.module_name}')

	def status(
		self, *, status: int, msg: str, args: Optional[Union[Dict[str, Any], DictObj]] = None
	) -> DictObj:
		if status != 200:
			logger.warning(
				f'BaseModule.status with msg \'{msg}\' is not called with \'200\' status code. Use BaseModule.exception instead.'
			)

		status_dict = {'status': status, 'msg': msg, 'args': {}}
		if args and type(args) == DictObj:
			if 'code' in args:
				args[
					'code'
				] = f'{self.package_name.upper()}_{self.module_name.upper()}_{args["code"]}'
			status_dict['args'] = args
		elif args and type(args) == dict:
			if 'code' in args.keys():
				args[
					'code'
				] = f'{self.package_name.upper()}_{self.module_name.upper()}_{args["code"]}'
			status_dict['args'] = args
		return DictObj(status_dict)

	def exception(
		self, *, status: int, msg: str, args: Optional[Union[Dict[str, Any], DictObj]] = None
	) -> MethodException:
		if status == 200:
			logger.warning(
				f'BaseModule.exception with msg \'{msg}\' is called with \'200\' status code. Use BaseModule.status instead.'
			)

		status_dict = {'status': status, 'msg': msg, 'args': {}}
		if args and type(args) == DictObj:
			if 'code' in args:
				args[
					'code'
				] = f'{self.package_name.upper()}_{self.module_name.upper()}_{args["code"]}'
			status_dict['args'] = args
		elif args and type(args) == dict:
			if 'code' in args.keys():
				args[
					'code'
				] = f'{self.package_name.upper()}_{self.module_name.upper()}_{args["code"]}'
			status_dict['args'] = args
		return MethodException(DictObj(status_dict))

	def __getattribute__(self, attr):
		# [DOC] Module is not yet initialised, skip to return exact attr
		try:
			object.__getattribute__(self, 'methods')
		except AttributeError:
			return object.__getattribute__(self, attr)
		# [DOC] Module is initialised attempt to check for methods
		if attr in object.__getattribute__(self, 'methods').keys():
			return object.__getattribute__(self, 'methods')[attr]
		elif attr.startswith('_method_'):
			return object.__getattribute__(self, attr.replace('_method_', ''))
		else:
			return object.__getattribute__(self, attr)

	async def pre_read(
		self,
		skip_events: NAWAH_EVENTS,
		env: NAWAH_ENV,
		query: Query,
		doc: NAWAH_DOC,
		payload: Dict[str, Any],
	) -> PRE_HANDLER_RETURN:
		return (skip_events, env, query, doc, payload)

	async def on_read(
		self,
		results: Dict[str, Any],
		skip_events: NAWAH_EVENTS,
		env: NAWAH_ENV,
		query: Query,
		doc: NAWAH_DOC,
		payload: Dict[str, Any],
	) -> ON_HANDLER_RETURN:
		return (results, skip_events, env, query, doc, payload)

	async def read(
		self,
		skip_events: NAWAH_EVENTS = [],
		env: NAWAH_ENV = {},
		query: Union[NAWAH_QUERY, Query] = [],
		doc: NAWAH_DOC = {},
	) -> DictObj:
		if not self.collection:
			raise self.exception(
				status=400,
				msg='Utility module can\'t call \'read\' method.',
				args={'code': 'INVALID_CALL'},
			)

		payload: Dict[str, Any] = {}

		query = cast(Query, query)

		if Event.PRE not in skip_events:
			pre_read = await self.pre_read(
				skip_events=skip_events, env=env, query=query, doc=doc, payload=payload
			)
			skip_events, env, query, doc, payload = pre_read

			# [DOC] Check if __results are passed in payload
			if '__results' in payload.keys():
				return payload['__results']

		# [DOC] Check for cache workflow instructins
		if self.cache:
			results: Optional[Dict[str, Any]] = None
			for cache_set in self.cache:
				if cache_set.condition(skip_events=skip_events, env=env, query=query) == True:
					cache_key = f'{str(query._query)}____{str(query._special)}'
					if Event.EXTN in skip_events:
						cache_key += '____EVENT_EXTN'
					if cache_key in cache_set.queries.keys():
						if cache_set.period:
							if (
								cache_set.queries[cache_key].query_time
								+ datetime.timedelta(seconds=cache_set.period)
							) < datetime.datetime.utcnow():
								if not results:
									results = await Data.read(
										env=env,
										collection_name=self.collection,
										attrs=self.attrs,
										query=query,
										skip_extn='$extn' in query or Event.EXTN in skip_events,
									)
								cache_set.queries[cache_key] = CACHED_QUERY(results=results)
							else:
								results = cache_set.queries[cache_key].results
								results['cache'] = cache_set.queries[cache_key].query_time.isoformat()
						else:
							results = cache_set.queries[cache_key].results
							results['cache'] = cache_set.queries[cache_key].query_time.isoformat()
					else:
						if not results:
							results = await Data.read(
								env=env,
								collection_name=self.collection,
								attrs=self.attrs,
								query=query,
								skip_extn='$extn' in query or Event.EXTN in skip_events,
							)
						cache_set.queries[cache_key] = CACHED_QUERY(results=results)
			if not results:
				results = await Data.read(
					env=env,
					collection_name=self.collection,
					attrs=self.attrs,
					query=query,
					skip_extn='$extn' in query or Event.EXTN in skip_events,
				)
		else:
			results = await Data.read(
				env=env,
				collection_name=self.collection,
				attrs=self.attrs,
				query=query,
				skip_extn='$extn' in query or Event.EXTN in skip_events,
			)
		if Event.ON not in skip_events:
			on_read = await self.on_read(
				results=results,
				skip_events=skip_events,
				env=env,
				query=query,
				doc=doc,
				payload=payload,
			)
			results, skip_events, env, query, doc, payload = on_read
			# [DOC] if $attrs query arg is present return only required keys.
			if '$attrs' in query:
				query['$attrs'].insert(0, '_id')
				for i in range(len(results['docs'])):
					results['docs'][i] = BaseModel(
						{
							attr: results['docs'][i][attr]
							for attr in query['$attrs']
							if attr in results['docs'][i]._attrs()
						}
					)

		return self.status(status=200, msg=f'Found {results["count"]} docs.', args=results)

	async def pre_watch(
		self,
		skip_events: NAWAH_EVENTS,
		env: NAWAH_ENV,
		query: Query,
		doc: NAWAH_DOC,
		payload: Dict[str, Any],
	) -> PRE_HANDLER_RETURN:
		return (skip_events, env, query, doc, payload)

	async def on_watch(
		self,
		results: Dict[str, Any],
		skip_events: NAWAH_EVENTS,
		env: NAWAH_ENV,
		query: Query,
		doc: NAWAH_DOC,
		payload: Dict[str, Any],
	) -> ON_HANDLER_RETURN:
		return (results, skip_events, env, query, doc, payload)

	async def watch(
		self,
		skip_events: NAWAH_EVENTS = [],
		env: NAWAH_ENV = {},
		query: Union[NAWAH_QUERY, Query] = [],
		doc: NAWAH_DOC = {},
	) -> AsyncGenerator[DictObj, DictObj]:
		if not self.collection:
			raise self.exception(
				status=400,
				msg='Utility module can\'t call \'watch\' method.',
				args={'code': 'INVALID_CALL'},
			)

		payload: Dict[str, Any] = {}

		query = cast(Query, query)

		if Event.PRE not in skip_events:
			pre_watch = await self.pre_watch(
				skip_events=skip_events, env=env, query=query, doc=doc, payload=payload
			)
			skip_events, env, query, doc, payload = pre_watch

		logger.debug('Preparing async loop at BaseModule')
		self.collection = cast(str, self.collection)
		async for results in Data.watch(
			env=env,
			collection_name=self.collection,
			attrs=self.attrs,
			query=query,
			skip_extn='$extn' in query or Event.EXTN in skip_events,
		):
			logger.debug(f'Received watch results at BaseModule: {results}')

			if 'stream' in results.keys():
				yield self.status(
					status=200, msg=f'Detected {results["count"]} docs.', args=results
				)
				continue

			if Event.ON not in skip_events:
				on_watch = await self.on_watch(
					results=results,
					skip_events=skip_events,
					env=env,
					query=query,
					doc=doc,
					payload=payload,
				)
				results, skip_events, env, query, doc, payload = on_watch

				# [DOC] if $attrs query arg is present return only required keys.
				if '$attrs' in query:
					query['$attrs'].insert(0, '_id')
					for i in range(len(results['docs'])):
						results['docs'][i] = BaseModel(
							{
								attr: results['docs'][i][attr]
								for attr in query['$attrs']
								if attr in results['docs'][i]._attrs()
							}
						)
			yield self.status(status=200, msg=f'Detected {results["count"]} docs.', args=results)

		logger.debug('Generator ended at BaseModule.')

	async def pre_create(
		self,
		skip_events: NAWAH_EVENTS,
		env: NAWAH_ENV,
		query: Query,
		doc: NAWAH_DOC,
		payload: Dict[str, Any],
	) -> PRE_HANDLER_RETURN:
		return (skip_events, env, query, doc, payload)

	async def on_create(
		self,
		results: Dict[str, Any],
		skip_events: NAWAH_EVENTS,
		env: NAWAH_ENV,
		query: Query,
		doc: NAWAH_DOC,
		payload: Dict[str, Any],
	) -> ON_HANDLER_RETURN:
		return (results, skip_events, env, query, doc, payload)

	async def create(
		self,
		skip_events: NAWAH_EVENTS = [],
		env: NAWAH_ENV = {},
		query: Union[NAWAH_QUERY, Query] = [],
		doc: NAWAH_DOC = {},
	) -> DictObj:
		if not self.collection:
			raise self.exception(
				status=400,
				msg='Utility module can\'t call \'create\' method.',
				args={'code': 'INVALID_CALL'},
			)

		payload: Dict[str, Any] = {}

		query = cast(Query, query)

		# [DOC] Check for __create_draft
		if '__create_draft' in query:
			# [DOC] Check for conflicts
			if len(doc.keys()):
				raise self.exception(
					status=400,
					msg='Can\'t use \'__create_draft\' with attrs provided in \'doc\'',
					args={'code': 'INVALID_CREATE_DRAFT_DOC'},
				)

			# [DOC] Load __create_draft and use it as doc
			draft_results = await Data.read(
				env=env,
				collection_name=self.collection,
				attrs=self.attrs,
				query=Query([{'_id': query['__create_draft'][0], '__create_draft': True}]),
				skip_process=True,
			)

			if not draft_results['count']:
				raise self.exception(
					status=400,
					msg='Invalid \'__create_draft\' doc.',
					args={'code': 'INVALID_CREATE_DRAFT'},
				)

			# [DOC] Data.read returns doc as BaseModel, extract values using BaseModel._attrs()
			doc = draft_results['docs'][0]._attrs()
			# [DOC] Delete special attrs
			del doc['_id']

		if Event.PRE not in skip_events:
			pre_create = await self.pre_create(
				skip_events=skip_events, env=env, query=query, doc=doc, payload=payload
			)
			skip_events, env, query, doc, payload = pre_create

			# [DOC] Check if __results are passed in payload
			if '__results' in payload.keys():
				return payload['__results']

		# [DOC] Expant dot-notated keys onto dicts
		doc = _expand_attr(doc=doc)
		# [DOC] Deleted all extra doc args
		doc = {
			attr: doc[attr]
			for attr in ['_id', '__create_draft', '__update_draft', *self.attrs.keys()]
			if attr in doc.keys() and doc[attr] != None
		}
		# [DOC] Append host_add, user_agent, create_time, diff if it's present in attrs.
		if (
			'user' in self.attrs.keys()
			and 'host_add' not in doc.keys()
			and env['session']
			and Event.ARGS not in skip_events
		):
			doc['user'] = env['session'].user._id
		if 'create_time' in self.attrs.keys():
			doc['create_time'] = datetime.datetime.utcnow().isoformat()
		if 'host_add' in self.attrs.keys() and 'host_add' not in doc.keys():
			doc['host_add'] = env['REMOTE_ADDR']
		if 'user_agent' in self.attrs.keys() and 'user_agent' not in doc.keys():
			doc['user_agent'] = env['HTTP_USER_AGENT']
		if Event.ARGS not in skip_events:
			mode: Literal['create', 'create_draft'] = 'create'
			if '__create_draft' in doc.keys() and doc['__create_draft'] == True:
				if not self.create_draft:
					raise self.exception(
						status=400,
						msg=f'Module \'{self.package_name.upper()}_{self.module_name.upper()}\' doesn\'t support \'create_draft\'',
						args={'code': 'NO_CREATE_DRAFT'},
					)
				if type(self.create_draft) == ATTR:
					self.create_draft = cast(ATTR, self.create_draft)
					# [DOC] Attr Type TYPE create_draft, call the funcion and catch InvalidAttrException
					try:
						await self.create_draft._args['func'](
							mode='create',
							attr_name='create_draft',
							attr_type=self.create_draft,
							attr_val=None,
							skip_events=skip_events,
							env=env,
							query=query,
							doc=doc,
							scope=doc,
						)
					except:
						raise self.exception(
							status=400,
							msg=f'Module \'{self.package_name.upper()}_{self.module_name.upper()}\' \'create_draft\' failed.',
							args={'code': 'NO_CREATE_DRAFT_CONDITION'},
						)

				mode = 'create_draft'

			elif '__update_draft' in doc.keys() and (
				type(doc['__update_draft']) == ObjectId
				or (
					type(doc['__update_draft']) == str
					and re.match(r'^[0-9a-fA-F]{24}$', doc['__update_draft'])
				)
			):
				doc['__update_draft'] = ObjectId(doc['__update_draft'])
				if not self.update_draft:
					raise self.exception(
						status=400,
						msg=f'Module \'{self.package_name.upper()}_{self.module_name.upper()}\' doesn\'t support \'update_draft\'',
						args={'code': 'NO_UPDATE_DRAFT'},
					)
				if type(self.update_draft) == ATTR:
					self.update_draft = cast(ATTR, self.update_draft)
					# [DOC] Attr Type TYPE update_draft, call the funcion and catch InvalidAttrException
					try:
						await self.update_draft._args['func'](
							mode='create',
							attr_name='update_draft',
							attr_type=self.update_draft,
							attr_val=None,
							skip_events=skip_events,
							env=env,
							query=query,
							doc=doc,
							scope=doc,
						)
					except:
						raise self.exception(
							status=400,
							msg=f'Module \'{self.package_name.upper()}_{self.module_name.upper()}\' \'update_draft\' failed.',
							args={'code': 'NO_UPDATE_DRAFT_CONDITION'},
						)

				mode = 'create_draft'
			# [DOC] Check presence and validate all attrs in doc args
			try:
				await validate_doc(
					mode=mode,
					doc=doc,
					attrs=self.attrs,
					skip_events=skip_events,
					env=env,
					query=query,
				)
			except MissingAttrException as e:
				raise self.exception(
					status=400,
					msg=f'{str(e)} for \'create\' request on module \'{self.package_name.upper()}_{self.module_name.upper()}\'.',
					args={'code': 'MISSING_ATTR'},
				)
			except InvalidAttrException as e:
				raise self.exception(
					status=400,
					msg=f'{str(e)} for \'create\' request on module \'{self.package_name.upper()}_{self.module_name.upper()}\'.',
					args={'code': 'INVALID_ATTR'},
				)
			except ConvertAttrException as e:
				raise self.exception(
					status=400,
					msg=f'{str(e)} for \'create\' request on module \'{self.package_name.upper()}_{self.module_name.upper()}\'.',
					args={'code': 'CONVERT_INVALID_ATTR'},
				)
			# [DOC] Check unique_attrs
			if self.unique_attrs:
				unique_attrs_query: List[Any] = [[]]
				for attr in self.unique_attrs:
					if type(attr) == str:
						attr = cast(str, attr)
						unique_attrs_query[0].append({attr: doc[attr]})
					elif type(attr) == tuple:
						unique_attrs_query[0].append({child_attr: doc[child_attr] for child_attr in attr})
					# [TODO] Implement use of single-item dict with LITERAL Attr Type for dynamic unique check based on doc value
				unique_attrs_query.append({'$limit': 1})
				unique_attrs_query = cast(NAWAH_QUERY, unique_attrs_query)
				unique_results = await self.read(
					skip_events=[Event.PERM], env=env, query=unique_attrs_query
				)
				if unique_results.args.count:
					unique_attrs_str = ', '.join(
						map(
							lambda _: ('(' + ', '.join(_) + ')') if type(_) == tuple else _,  # type: ignore
							self.unique_attrs,
						)
					)
					raise self.exception(
						status=400,
						msg=f'A doc with the same \'{unique_attrs_str}\' already exists.',
						args={'code': 'DUPLICATE_DOC'},
					)
		# [DOC] Execute Data driver create
		results = await Data.create(
			env=env, collection_name=self.collection, attrs=self.attrs, doc=doc
		)

		# [DOC] Check for __create_draft and delete it
		if '__create_draft' in query:
			delete_draft_results = await Data.delete(
				env=env,
				collection_name=self.collection,
				attrs=self.attrs,
				docs=[ObjectId(query['__create_draft'][0])],
				strategy=DELETE_STRATEGY.FORCE_SYS,
			)
			if delete_draft_results['count'] != 1:
				logger.error(
					f'Failed to delete \'__create_draft\'. Results: {delete_draft_results}'
				)

		if Event.ON not in skip_events:
			on_create = await self.on_create(
				results=results,
				skip_events=skip_events,
				env=env,
				query=query,
				doc=doc,
				payload=payload,
			)
			results, skip_events, env, query, doc, payload = on_create

		# [DOC] create soft action is to only return the new created doc _id.
		if Event.SOFT in skip_events:
			read_results = await self.read(
				skip_events=[Event.PERM], env=env, query=[[{'_id': results['docs'][0]}]]
			)
			results = read_results.args

		# [DOC] Module collection is updated, update_cache
		asyncio.create_task(self.update_cache(env=env))

		return self.status(status=200, msg=f'Created {results["count"]} docs.', args=results)

	async def pre_update(
		self,
		skip_events: NAWAH_EVENTS,
		env: NAWAH_ENV,
		query: Query,
		doc: NAWAH_DOC,
		payload: Dict[str, Any],
	) -> PRE_HANDLER_RETURN:
		return (skip_events, env, query, doc, payload)

	async def on_update(
		self,
		results: Dict[str, Any],
		skip_events: NAWAH_EVENTS,
		env: NAWAH_ENV,
		query: Query,
		doc: NAWAH_DOC,
		payload: Dict[str, Any],
	) -> ON_HANDLER_RETURN:
		return (results, skip_events, env, query, doc, payload)

	async def update(
		self,
		skip_events: NAWAH_EVENTS = [],
		env: NAWAH_ENV = {},
		query: Union[NAWAH_QUERY, Query] = [],
		doc: NAWAH_DOC = {},
	) -> DictObj:
		if not self.collection:
			raise self.exception(
				status=400,
				msg='Utility module can\'t call \'update\' method.',
				args={'code': 'INVALID_CALL'},
			)

		payload: Dict[str, Any] = {}

		query = cast(Query, query)

		update_draft = None
		# [DOC] Check for __update_draft
		if '__update_draft' in query:
			update_draft = query['__update_draft'][0]
			# [DOC] Check for conflicts
			if len(doc.keys()):
				raise self.exception(
					status=400,
					msg='Can\'t use \'__update_draft\' with attrs provided in \'doc\'',
					args={'code': 'INVALID_UPDATE_DRAFT_DOC'},
				)

			# [DOC] Load __create_draft and use it as doc
			draft_results = await Data.read(
				env=env,
				collection_name=self.collection,
				attrs=self.attrs,
				query=Query(
					[{'_id': query['__update_draft'][0], '__update_draft': {'$ne': False}}]
				),
				skip_process=True,
			)

			if not draft_results['count']:
				raise self.exception(
					status=400,
					msg='Invalid \'__update_draft\' doc.',
					args={'code': 'INVALID_UPDATE_DRAFT'},
				)

			# [DOC] Update query per the value of doc.__update_draft
			del query['__update_draft'][0]
			query.append({'_id': draft_results['docs'][0]['__update_draft']})
			# [DOC] Data.read returns doc as BaseModel, extract values using BaseModel._attrs()
			doc = draft_results['docs'][0]._attrs()
			# [DOC] Delete special attrs
			del doc['_id']

		if Event.PRE not in skip_events:
			pre_update = await self.pre_update(
				skip_events=skip_events, env=env, query=query, doc=doc, payload=payload
			)
			skip_events, env, query, doc, payload = pre_update

			# [DOC] Check if __results are passed in payload
			if '__results' in payload.keys():
				return payload['__results']

		# [DOC] Check presence and validate all attrs in doc args
		try:
			await validate_doc(
				mode='update',
				doc=doc,
				attrs=self.attrs,
				skip_events=skip_events,
				env=env,
				query=query,
			)
		except MissingAttrException as e:
			raise self.exception(
				status=400,
				msg=f'{str(e)} for \'update\' request on module \'{self.package_name.upper()}_{self.module_name.upper()}\'.',
				args={'code': 'MISSING_ATTR'},
			)
		except InvalidAttrException as e:
			raise self.exception(
				status=400,
				msg=f'{str(e)} for \'update\' request on module \'{self.package_name.upper()}_{self.module_name.upper()}\'.',
				args={'code': 'INVALID_ATTR'},
			)
		except ConvertAttrException as e:
			raise self.exception(
				status=400,
				msg=f'{str(e)} for \'update\' request on module \'{self.package_name.upper()}_{self.module_name.upper()}\'.',
				args={'code': 'CONVERT_INVALID_ATTR'},
			)
		# [DOC] Delete all attrs not belonging to the module
		shadow_doc = {}
		for attr_name in ['_id', *doc.keys()]:
			attr_root = attr_name.split('.')[0].split(':')[0]
			# [DOC] Check top-level attr belong to module
			if attr_root not in self.attrs.keys():
				continue
			# [DOC] Check attr is valid Doc Oper
			if (
				type(doc[attr_name]) == dict
				and doc[attr_name].keys()
				and list(doc[attr_name].keys())[0][0] == '$'
				and doc[attr_name][list(doc[attr_name].keys())[0]] == None
			):
				continue
			# [DOC] Add non-None attrs to shadow_doc
			if doc[attr_name] != None:
				shadow_doc[attr_name] = doc[attr_name]

		# [DOC] Replace all NAWAH_VALUES.NONE_VALUE with None
		_set_none_value(doc=shadow_doc)
		doc = shadow_doc

		# [DOC] Check if there is anything yet to update
		if not len(doc.keys()):
			return self.status(status=200, msg='Nothing to update.', args={})
		# [DOC] Find which docs are to be updated
		docs_results = await Data.read(
			env=env,
			collection_name=self.collection,
			attrs=self.attrs,
			query=query,
			skip_process=True,
		)
		# [DOC] Check unique_attrs
		if self.unique_attrs:
			# [DOC] If any of the unique_attrs is present in doc, and docs_results is > 1, we have duplication
			if len(docs_results['docs']) > 1:
				unique_attrs_check = True
				for attr in self.unique_attrs:
					if type(attr) == str and attr in doc.keys():
						unique_attrs_check = False
						break
					elif type(attr) == tuple:
						for child_attr in attr:
							if not unique_attrs_check:
								break
							if child_attr in doc.keys():
								unique_attrs_check = False
								break

				if not unique_attrs_check:
					raise self.exception(
						status=400,
						msg='Update call query has more than one doc as results. This would result in duplication.',
						args={'code': 'MULTI_DUPLICATE'},
					)

			# [DOC] Check if any of the unique_attrs are present in doc
			# [TODO] Replace with any
			if sum(1 for attr in doc.keys() if attr in self.unique_attrs) > 0:
				# [DOC] Check if the doc would result in duplication after update
				unique_attrs_query: List[Any] = [[]]
				for attr in self.unique_attrs:
					if type(attr) == str:
						attr = cast(str, attr)
						if attr in doc.keys():
							unique_attrs_query[0].append({attr: doc[attr]})
					elif type(attr) == tuple:
						unique_attrs_query[0].append(
							{child_attr: doc[child_attr] for child_attr in attr if attr in doc.keys()}
						)
				unique_attrs_query.append(
					{'_id': {'$nin': [doc._id for doc in docs_results['docs']]}}
				)
				unique_attrs_query.append({'$limit': 1})
				unique_attrs_query = cast(NAWAH_QUERY, unique_attrs_query)
				unique_results = await self.read(
					skip_events=[Event.PERM], env=env, query=unique_attrs_query
				)
				if unique_results.args.count:
					unique_attrs_str = ', '.join(
						map(
							lambda _: ('(' + ', '.join(_) + ')') if type(_) == tuple else _,  # type: ignore
							self.unique_attrs,
						)
					)
					raise self.exception(
						status=400,
						msg=f'A doc with the same \'{unique_attrs_str}\' already exists.',
						args={'code': 'DUPLICATE_DOC'},
					)
		results = await Data.update(
			env=env,
			collection_name=self.collection,
			attrs=self.attrs,
			docs=[doc._id for doc in docs_results['docs']],
			doc=doc,
		)

		# [DOC] Check for update_draft and delete it
		if update_draft:
			delete_draft_results = await Data.delete(
				env=env,
				collection_name=self.collection,
				attrs=self.attrs,
				docs=[ObjectId(update_draft)],
				strategy=DELETE_STRATEGY.FORCE_SYS,
			)
			if delete_draft_results['count'] != 1:
				logger.error(
					f'Failed to delete \'__update_draft\'. Results: {delete_draft_results}'
				)

		if Event.ON not in skip_events:
			on_update = await self.on_update(
				results=results,
				skip_events=skip_events,
				env=env,
				query=query,
				doc=doc,
				payload=payload,
			)
			results, skip_events, env, query, doc, payload = on_update

		# [DOC] If at least one doc updated, and module has diff enabled, and __DIFF__ not skipped:
		if results['count'] and self.diff and Event.DIFF not in skip_events:
			if type(self.diff) == ATTR:
				# [DOC] # [DOC] Attr Type TYPE diff, call the funcion and catch InvalidAttrException
				self.diff = cast(ATTR, self.diff)
				try:
					await self.diff._args['func'](
						mode='create',
						attr_name='diff',
						attr_type=self.diff,
						attr_val=None,
						skip_events=skip_events,
						env=env,
						query=query,
						doc=doc,
						scope=doc,
					)

					# [DOC] if function passes, create Diff doc with default callable
					diff_vars = doc
					diff_results = await Config.modules['diff'].create(
						skip_events=[Event.PERM],
						env=env,
						query=query,
						doc={'module': self.module_name, 'vars': diff_vars},
					)
					if diff_results.status != 200:
						logger.error(f'Failed to create Diff doc, results: {diff_results}')
				except:
					logger.debug(f'Skipped Diff Workflow due to failed condition.')
			else:
				diff_results = await Config.modules['diff'].create(
					skip_events=[Event.PERM],
					env=env,
					query=query,
					doc={'module': self.module_name, 'vars': doc},
				)
				if diff_results.status != 200:
					logger.error(f'Failed to create Diff doc, results: {diff_results}')
		else:
			logger.debug(
				f'Skipped Diff Workflow due to: {results["count"]}, {self.diff}, {Event.DIFF not in skip_events}'
			)

		# [DOC] Module collection is updated, update_cache
		asyncio.create_task(self.update_cache(env=env))

		return self.status(status=200, msg=f'Updated {results["count"]} docs.', args=results)

	async def pre_delete(
		self,
		skip_events: NAWAH_EVENTS,
		env: NAWAH_ENV,
		query: Query,
		doc: NAWAH_DOC,
		payload: Dict[str, Any],
	) -> PRE_HANDLER_RETURN:
		return (skip_events, env, query, doc, payload)

	async def on_delete(
		self,
		results: Dict[str, Any],
		skip_events: NAWAH_EVENTS,
		env: NAWAH_ENV,
		query: Query,
		doc: NAWAH_DOC,
		payload: Dict[str, Any],
	) -> ON_HANDLER_RETURN:
		return (results, skip_events, env, query, doc, payload)

	async def delete(
		self,
		skip_events: NAWAH_EVENTS = [],
		env: NAWAH_ENV = {},
		query: Union[NAWAH_QUERY, Query] = [],
		doc: NAWAH_DOC = {},
	) -> DictObj:
		if not self.collection:
			raise self.exception(
				status=400,
				msg='Utility module can\'t call \'delete\' method.',
				args={'code': 'INVALID_CALL'},
			)

		payload: Dict[str, Any] = {}

		query = cast(Query, query)

		if Event.PRE not in skip_events:
			pre_delete = await self.pre_delete(
				skip_events=skip_events, env=env, query=query, doc=doc, payload=payload
			)
			skip_events, env, query, doc, payload = pre_delete

			# [DOC] Check if __results are passed in payload
			if '__results' in payload.keys():
				return payload['__results']

		# [TODO]: confirm all extns are not linked.
		# [DOC] Pick delete strategy based on skip_events
		strategy = DELETE_STRATEGY.SOFT_SKIP_SYS
		if Event.SOFT not in skip_events and Event.SYS_DOCS in skip_events:
			strategy = DELETE_STRATEGY.SOFT_SYS
		elif Event.SOFT in skip_events:
			if Event.SYS_DOCS not in skip_events:
				strategy = DELETE_STRATEGY.FORCE_SKIP_SYS
			else:
				strategy = DELETE_STRATEGY.FORCE_SYS

		docs_results = results = await Data.read(
			env=env,
			collection_name=self.collection,
			attrs=self.attrs,
			query=query,
			skip_process=True,
		)
		results = await Data.delete(
			env=env,
			collection_name=self.collection,
			attrs=self.attrs,
			docs=[doc._id for doc in docs_results['docs']],
			strategy=strategy,
		)
		if Event.ON not in skip_events:
			on_delete = await self.on_delete(
				results=results,
				skip_events=skip_events,
				env=env,
				query=query,
				doc=doc,
				payload=payload,
			)
			results, skip_events, env, query, doc, payload = on_delete

		# [DOC] Module collection is updated, update_cache
		asyncio.create_task(self.update_cache(env=env))

		return self.status(status=200, msg=f'Deleted {results["count"]} docs.', args=results)

	async def pre_create_file(
		self,
		skip_events: NAWAH_EVENTS,
		env: NAWAH_ENV,
		query: Query,
		doc: NAWAH_DOC,
		payload: Dict[str, Any],
	) -> PRE_HANDLER_RETURN:
		return (skip_events, env, query, doc, payload)

	async def on_create_file(
		self,
		results: Dict[str, Any],
		skip_events: NAWAH_EVENTS,
		env: NAWAH_ENV,
		query: Query,
		doc: NAWAH_DOC,
		payload: Dict[str, Any],
	) -> ON_HANDLER_RETURN:
		return (results, skip_events, env, query, doc, payload)

	async def create_file(
		self,
		skip_events: NAWAH_EVENTS = [],
		env: NAWAH_ENV = {},
		query: Union[NAWAH_QUERY, Query] = [],
		doc: NAWAH_DOC = {},
	) -> DictObj:
		if not self.collection:
			raise self.exception(
				status=400,
				msg='Utility module can\'t call \'create_file\' method.',
				args={'code': 'INVALID_CALL'},
			)

		payload: Dict[str, Any] = {}

		query = cast(Query, query)

		if Event.PRE not in skip_events:
			pre_create_file = await self.pre_create_file(
				skip_events=skip_events, env=env, query=query, doc=doc, payload=payload
			)
			skip_events, env, query, doc, payload = pre_create_file

			# [DOC] Check if __results are passed in payload
			if '__results' in payload.keys():
				return payload['__results']

		# [TODO] Allow use dot-notated attr path in attr query attr
		if (
			query['attr'][0] not in self.attrs.keys()
			or type(self.attrs[query['attr'][0]]._type) != 'LIST'
			or not self.attrs[query['attr'][0]]._args['list'][0]._type != 'FILE'
		):
			raise self.exception(
				status=400, msg='Attr is invalid.', args={'code': 'INVALID_ATTR'}
			)

		update_results = await self.update(
			skip_events=[Event.PERM],
			env=env,
			query=[{'_id': query['_id'][0]}],
			doc={query['attr'][0]: {'$append': doc['file']}},
		)
		results = update_results['args']

		if Event.ON not in skip_events:
			on_create_file = await self.on_create_file(
				results=results,
				skip_events=skip_events,
				env=env,
				query=query,
				doc=doc,
				payload=payload,
			)
			results, skip_events, env, query, doc, payload = on_create_file

		return self.status(status=200, msg=f'Updated {results["count"]} docs.', args=results)

	async def pre_delete_file(
		self,
		skip_events: NAWAH_EVENTS,
		env: NAWAH_ENV,
		query: Query,
		doc: NAWAH_DOC,
		payload: Dict[str, Any],
	) -> PRE_HANDLER_RETURN:
		return (skip_events, env, query, doc, payload)

	async def on_delete_file(
		self,
		results: Dict[str, Any],
		skip_events: NAWAH_EVENTS,
		env: NAWAH_ENV,
		query: Query,
		doc: NAWAH_DOC,
		payload: Dict[str, Any],
	) -> ON_HANDLER_RETURN:
		return (results, skip_events, env, query, doc, payload)

	async def delete_file(
		self,
		skip_events: NAWAH_EVENTS = [],
		env: NAWAH_ENV = {},
		query: Union[NAWAH_QUERY, Query] = [],
		doc: NAWAH_DOC = {},
	) -> DictObj:
		if not self.collection:
			raise self.exception(
				status=400,
				msg='Utility module can\'t call \'delete_file\' method.',
				args={'code': 'INVALID_CALL'},
			)

		payload: Dict[str, Any] = {}

		query = cast(Query, query)

		if Event.PRE not in skip_events:
			pre_delete_file = await self.pre_delete_file(
				skip_events=skip_events, env=env, query=query, doc=doc, payload=payload
			)
			skip_events, env, query, doc, payload = pre_delete_file

			# [DOC] Check if __results are passed in payload
			if '__results' in payload.keys():
				return payload['__results']

		# [TODO] Allow use dot-notated attr path in attr query attr
		if (
			query['attr'][0] not in self.attrs.keys()
			or type(self.attrs[query['attr'][0]]._type) != 'LIST'
			or not self.attrs[query['attr'][0]]._args['list'][0]._type != 'FILE'
		):
			raise self.exception(
				status=400, msg='Attr is invalid.', args={'code': 'INVALID_ATTR'}
			)

		read_results = await self.read(
			skip_events=[Event.PERM], env=env, query=[{'_id': query['_id'][0]}]
		)
		if not read_results.args.count:
			raise self.exception(status=400, msg='Doc is invalid.', args={'code': 'INVALID_DOC'})
		doc = read_results.args.docs[0]

		if query['attr'][0] not in doc or not doc[query['attr'][0]]:
			raise self.exception(
				status=400,
				msg='Doc attr is invalid.',
				args={'code': 'INVALID_DOC_ATTR'},
			)

		if query['index'][0] not in range(len(doc[query['attr'][0]])):
			raise self.exception(
				status=400, msg='Index is invalid.', args={'code': 'INVALID_INDEX'}
			)

		if (
			type(doc[query['attr'][0]][query['index'][0]]) != dict
			or 'name' not in doc[query['attr'][0]][query['index'][0]].keys()
		):
			raise self.exception(
				status=400,
				msg='Index value is invalid.',
				args={'code': 'INVALID_INDEX_VALUE'},
			)

		if doc[query['attr'][0]][query['index'][0]]['name'] != query['name'][0]:  # type: ignore
			raise self.exception(
				status=400,
				msg='File name in query doesn\'t match value.',
				args={'code': 'FILE_NAME_MISMATCH'},
			)

		update_results = await self.update(
			skip_events=[Event.PERM],
			env=env,
			query=[{'_id': query['_id'][0]}],
			doc={query['attr'][0]: {'$del_val': [doc[query['attr'][0]][query['index'][0]]]}},
		)
		results = update_results['args']

		if Event.ON not in skip_events:
			on_delete_file = await self.on_delete_file(
				results=results,
				skip_events=skip_events,
				env=env,
				query=query,
				doc=doc,
				payload=payload,
			)
			results, skip_events, env, query, doc, payload = on_delete_file

		return results

	async def pre_retrieve_file(
		self,
		skip_events: NAWAH_EVENTS,
		env: NAWAH_ENV,
		query: Query,
		doc: NAWAH_DOC,
		payload: Dict[str, Any],
	) -> PRE_HANDLER_RETURN:
		return (skip_events, env, query, doc, payload)

	async def on_retrieve_file(
		self,
		results: Dict[str, Any],
		skip_events: NAWAH_EVENTS,
		env: NAWAH_ENV,
		query: Query,
		doc: NAWAH_DOC,
		payload: Dict[str, Any],
	) -> ON_HANDLER_RETURN:
		return (results, skip_events, env, query, doc, payload)

	async def retrieve_file(
		self,
		skip_events: NAWAH_EVENTS = [],
		env: NAWAH_ENV = {},
		query: Union[NAWAH_QUERY, Query] = [],
		doc: NAWAH_DOC = {},
	) -> DictObj:
		if not self.collection:
			raise self.exception(
				status=400,
				msg='Utility module can\'t call \'retrieve_file\' method.',
				args={'code': 'INVALID_CALL'},
			)

		payload: Dict[str, Any] = {}

		query = cast(Query, query)

		if Event.PRE not in skip_events:
			pre_retrieve_file = await self.pre_retrieve_file(
				skip_events=skip_events, env=env, query=query, doc=doc, payload=payload
			)
			skip_events, env, query, doc, payload = pre_retrieve_file

			# [DOC] Check if __results are passed in payload
			if '__results' in payload.keys():
				return payload['__results']

		attr_name = query['attr'][0]
		filename = query['filename'][0]
		if 'thumb' in query:
			thumb_dims: Optional[List[int]] = [int(dim) for dim in query['thumb'][0].split('x')]
		else:
			thumb_dims = None

		read_results = await self.read(
			skip_events=[Event.PERM] + skip_events,
			env=env,
			query=[{'_id': query['_id'][0]}],
		)
		if not read_results.args.count:
			raise self.exception(
				status=400,
				msg='File not found.',
				args={'code': 'NOT_FOUND', 'return': 'json'},
			)
		attr: Union[List[Dict[str, Any]], Dict[str, Any]]
		doc = read_results.args.docs[0]
		try:
			attr_path = attr_name.split('.')
			attr = doc
			for path in attr_path:
				attr = doc[path]
		except:
			raise self.exception(
				status=404,
				msg='File not found.',
				args={'code': 'NOT_FOUND', 'return': 'json'},
			)

		retrieved_file = None

		if type(attr) == list:
			attr = cast(List[Dict[str, Any]], attr)
			for item in attr:
				if item['name'] == filename:
					retrieved_file = item
					break
		elif type(attr) == dict:
			attr = cast(Dict[str, Any], attr)
			if attr['name'] == filename:
				retrieved_file = attr

		if not retrieved_file:
			# [DOC] No filename match
			raise self.exception(
				status=404,
				msg='File not found.',
				args={'code': 'NOT_FOUND', 'return': 'json'},
			)

		# [DOC] filematch!
		results: Dict[str, Any]
		results = {
			'docs': [
				DictObj(
					{
						'_id': query['_id'][0],
						'name': retrieved_file['name'],
						'type': retrieved_file['type'],
						'lastModified': retrieved_file['lastModified'],
						'size': retrieved_file['size'],
						'content': retrieved_file['content'],
					}
				)
			]
		}

		if thumb_dims:
			if retrieved_file['type'].split('/')[0] != 'image':
				raise self.exception(
					status=400,
					msg='File is not of type image to create thumbnail for.',
					args={'code': 'NOT_IMAGE', 'return': 'json'},
				)
			try:
				image = Image.open(io.BytesIO(retrieved_file['content']))
				image.thumbnail(thumb_dims)
				stream = io.BytesIO()
				image.save(stream, format=image.format)
				stream.seek(0)
				results['docs'][0]['content'] = stream.read()
			except:
				pass

		if Event.ON not in skip_events:
			on_retrieve_file = await self.on_retrieve_file(
				results=results,
				skip_events=skip_events,
				env=env,
				query=query,
				doc=doc,
				payload=payload,
			)
			results, skip_events, env, query, doc, payload = on_retrieve_file

		results['return'] = 'file'
		return self.status(status=200, msg='File attached to response.', args=results)

	async def update_cache(
		self,
		skip_events: NAWAH_EVENTS = [],
		env: NAWAH_ENV = {},
		query: Union[NAWAH_QUERY, Query] = [],
		doc: NAWAH_DOC = {},
	) -> DictObj:
		if self.collection and self.cache:
			for cache_set in self.cache:
				# [DOC] Check if Cache Set requires recreating queries, or dismiss them
				if cache_set.cache_strategy == CACHE_STRATEGY.DISMISS:
					cache_set.queries = {}
					continue

				for cache_key in cache_set.queries.keys():
					del cache_set.queries[cache_key]
					cache_key_split = cache_key.split('____')
					cache_query: NAWAH_QUERY = eval(cache_key_split[0])
					cache_special: NAWAH_QUERY = eval(cache_key_split[1])
					cache_query.append(cache_special)
					results = await Data.read(
						env=env,
						collection_name=self.collection,
						attrs=self.attrs,
						query=Query(cache_query),
						skip_extn=(len(cache_key_split) == 3 and cache_key_split[2] == 'EVENT_EXTN'),
					)
					cache_set.queries[cache_key] = CACHED_QUERY(results=results)
		return self.status(status=200, msg='Cache deleted.', args={})



from typing import (
	Literal,
	Any,
	Dict,
	List,
	Union,
	TypedDict,
	cast,
	Optional,
	TYPE_CHECKING,
)

import copy

from ._attr import SPECIAL_ATTRS
from ._dictobj import DictObj
from ._exceptions import InvalidQueryArgException, UnknownQueryArgException

if TYPE_CHECKING:
	from ._types import NAWAH_QUERY, NAWAH_QUERY_SPECIAL


QUERY_INDEX_RECORD = TypedDict(
	'QUERY_INDEX_RECORD', {'oper': str, 'path': str, 'val': Any}
)


class Query(list):
	_query: 'NAWAH_QUERY'
	_special: 'NAWAH_QUERY_SPECIAL'
	_index: Dict[str, List[QUERY_INDEX_RECORD]]

	def __init__(self, query: Union['NAWAH_QUERY', 'Query']):
		self._query = query
		if type(self._query) == Query:
			query = cast(Query, query)
			self._query = query._query + [query._special]
		self._special = {}
		self._index: Dict[str, List[QUERY_INDEX_RECORD]] = {}
		self._create_index(self._query)
		super().__init__(self._query)

	def __repr__(self):
		return str(self._query + [self._special])

	def _create_index(self, query: 'NAWAH_QUERY', path=[]):
		if not path:
			self._index: Dict[str, List[QUERY_INDEX_RECORD]] = {}
		for i in range(len(query)):
			if type(query[i]) == dict:
				del_attrs = []
				for attr in query[i].keys():
					if attr in SPECIAL_ATTRS.__args__:  # type: ignore
						self._special[attr] = query[i][attr]  # type: ignore
						del_attrs.append(attr)
					elif attr.startswith('__or'):
						self._create_index(query[i][attr], path=path + [i, attr])  # type: ignore
					else:
						if (
							type(query[i][attr]) == dict  # type: ignore
							and len(query[i][attr].keys()) == 1  # type: ignore
							and list(query[i][attr].keys())[0][0] == '$'  # type: ignore
						):
							attr_oper = list(query[i][attr].keys())[0]  # type: ignore
						else:
							attr_oper = '$eq'
						if attr not in self._index.keys():
							self._index[attr] = []
						if isinstance(query[i][attr], DictObj):  # type: ignore
							query[i][attr] = query[i][attr]._id  # type: ignore
						Query.validate_arg(arg_name=attr, arg_oper=attr_oper, arg_val=query[i][attr])  # type: ignore
						self._index[attr].append(
							{
								'oper': attr_oper,
								'path': path + [i],
								'val': query[i][attr],  # type: ignore
							}
						)
				for attr in del_attrs:
					del query[i][attr]  # type: ignore
			elif type(query[i]) == list:
				self._create_index(query[i], path=path + [i])  # type: ignore
		if not path:
			self._query = self._sanitise_query()

	def _sanitise_query(self, query: 'NAWAH_QUERY' = None):
		if query == None:
			query = self._query
		query = cast('NAWAH_QUERY', query)
		query_shadow = []
		for step in query:
			if type(step) == dict:
				for attr in step.keys():
					if attr.startswith('__or'):
						step[attr] = self._sanitise_query(step[attr])  # type: ignore
						if len(step[attr]):  # type: ignore
							query_shadow.append(step)
							break
					elif attr[0] != '$':
						query_shadow.append(step)
						break
			elif type(step) == list:
				step = self._sanitise_query(step)  # type: ignore
				if len(step):
					query_shadow.append(step)
		return query_shadow

	def __deepcopy__(self, memo):
		return Query(copy.deepcopy(self._query + [self._special]))

	def append(self, obj: Any):
		self._query.append(obj)
		self._create_index(self._query)
		super().__init__(self._query)

	def __contains__(self, attr: str):  # type: ignore
		if attr in SPECIAL_ATTRS.__args__:  # type: ignore
			return attr in self._special.keys()
		else:
			if ':' in attr:
				attr_index, attr_oper = attr.split(':')
			else:
				attr_index = attr
				attr += ':$eq'
				attr_oper = '$eq'

			if attr_index in self._index.keys():
				for val in self._index[attr_index]:
					if val['oper'] == attr_oper or attr_oper == '*':
						return True
			return False

	def __getitem__(self, attr: Union[SPECIAL_ATTRS, str]):  # type: ignore
		if attr in SPECIAL_ATTRS.__args__:  # type: ignore
			return self._special[attr]  # type: ignore
		else:
			attrs = []
			vals = []
			paths: List[List[int]] = []
			indexes = []
			attr_filter: Optional[str] = None
			oper_filter: Optional[str] = None

			if attr.split(':')[0] != '*':
				attr_filter = attr.split(':')[0]

			if ':' not in attr:
				oper_filter = '$eq'
				attr += ':$eq'
			elif ':*' not in attr:
				oper_filter = attr.split(':')[1]

			for index_attr in self._index.keys():
				if attr_filter and index_attr != attr_filter:
					continue
				i = 0
				for val in self._index[index_attr]:
					if not oper_filter or (oper_filter and val['oper'] == oper_filter):
						attrs.append(index_attr)
						# [TODO] Simplify this condition by enforcing Query Args with $eq Query Oper are always stripped down to value
						if not oper_filter or (
							oper_filter == '$eq'
							and (type(val['val']) != dict or '$eq' not in val['val'].keys())
						):
							vals.append(val['val'])
						else:
							vals.append(val['val'][oper_filter])
						paths.append(val['path'])
						indexes.append(i)
						i += 1
			return QueryAttrList(self, attrs, paths, indexes, vals)

	def __setitem__(self, attr: str, val: Any):  # type: ignore
		if attr[0] != '$':
			raise Exception('Non-special attrs can only be updated by attr index.')
		self._special[attr] = val  # type: ignore

	def __delitem__(self, attr: str):  # type: ignore
		if attr[0] != '$':
			raise Exception('Non-special attrs can only be deleted by attr index.')
		del self._special[attr]  # type: ignore

	@classmethod
	def validate_arg(cls, *, arg_name, arg_oper, arg_val):
		if arg_oper in ['$ne', '$eq']:
			return
		elif arg_oper in ['$gt', '$gte', '$lt', '$lte']:
			if type(arg_val[arg_oper]) not in [str, int, float]:
				raise InvalidQueryArgException(
					arg_name=arg_name,
					arg_oper=arg_oper,
					arg_type=[str, int, float],
					arg_val=arg_val[arg_oper],
				)
		elif arg_oper == '$bet':
			if (
				type(arg_val[arg_oper]) != list
				or len(arg_val[arg_oper]) != 2
				or type(arg_val[arg_oper][0]) not in [str, int, float]
				or type(arg_val[arg_oper][1]) not in [str, int, float]
			):
				raise InvalidQueryArgException(
					arg_name=arg_name,
					arg_oper=arg_oper,
					arg_type=list,
					arg_val=arg_val[arg_oper],
				)
		elif arg_oper in ['$all', '$in', '$nin']:
			if type(arg_val[arg_oper]) != list:
			# [TODO] Validate why this was added
			# or not len(arg_val[arg_oper]):
				raise InvalidQueryArgException(
					arg_name=arg_name,
					arg_oper=arg_oper,
					arg_type=list,
					arg_val=arg_val[arg_oper],
				)
		elif arg_oper == '$regex':
			if type(arg_val[arg_oper]) != str:
				raise InvalidQueryArgException(
					arg_name=arg_name,
					arg_oper=arg_oper,
					arg_type=str,
					arg_val=arg_val[arg_oper],
				)
		else:
			raise UnknownQueryArgException(arg_name=arg_name, arg_oper=arg_oper)


class QueryAttrList(list):
	def __init__(
		self,
		query: Query,
		attrs: List[str],
		paths: List[List[int]],
		indexes: List[int],
		vals: List[Any],
	):
		self._query = query
		self._attrs = attrs
		self._paths = paths
		self._indexes = indexes
		self._vals = vals
		super().__init__(vals)

	def __setitem__(self, item: Union[Literal['*'], int], val: Any):  # type: ignore
		if item == '*':
			for i in range(len(self._vals)):
				self.__setitem__(i, val)
		else:
			instance_attr = self._query._query
			for path_part in self._paths[item]:
				instance_attr = instance_attr[path_part]  # type: ignore
			instance_attr[self._attrs[item].split(':')[0]] = val  # type: ignore
			self._query._create_index(self._query._query)

	def __delitem__(self, item: Union[Literal['*'], int]):  # type: ignore
		if item == '*':
			for i in range(len(self._vals)):
				self.__delitem__(i)
		else:
			instance_attr = self._query._query
			for path_part in self._paths[item]:
				instance_attr = instance_attr[path_part]  # type: ignore
			del instance_attr[self._attrs[item].split(':')[0]]  # type: ignore
			self._query._create_index(self._query._query)

	def replace_attr(self, item: Union[Literal['*'], int], new_attr: str):
		if item == '*':
			for i in range(len(self._vals)):
				self.replace_attr(i, new_attr)
		else:
			instance_attr = self._query._query
			for path_part in self._paths[item]:
				instance_attr = instance_attr[path_part]  # type: ignore
			# [DOC] Set new attr
			instance_attr[new_attr] = instance_attr[self._attrs[item].split(':')[0]]  # type: ignore
			# [DOC] Delete old attr
			del instance_attr[self._attrs[item].split(':')[0]]  # type: ignore
			# [DOC] Update index
			self._query._create_index(self._query._query)